#! /usr/local/env python3
import argparse
import glob
import logging
import multiprocessing as mp
import os
import re
from typing import Optional, List

import matplotlib.style as style
import matplotlib.pyplot as plt
import matplotlib.patches as patches
import numpy as np
import pandas as pd
from scipy.stats import ttest_ind, ttest_rel
import statsmodels.api as sm
import statsmodels.formula.api as smf

import traja
from traja import TrajaDataFrame

try:
    import seaborn as sns
except ImportError:
    raise ImportError(
        "##########################\n"
        "seaborn is not installed. \n"
        "install it with\n"
        "  pip install seaborn\n"
        "\n"
        "##########################"
    )

DARK = "Dark"
LIGHT = "Light"
DAYS_COL = "days_from_surgery"


def glm(
    daily: pd.DataFrame,
    response_var: str = "displacement",
    cov_struct: Optional[str] = None,
    rmAnova=False,
    verbose=False,
):
    """Generalized linear model for daily response."""
    kwargs = {}

    if cov_struct == "auto":
        cov_struct = sm.cov_struct.Autoregressive()
        kwargs["cov_struct"] = cov_struct
    elif cov_struct == "exchangable":
        cov_struct = sm.cov_struct.Exchangeable()
        kwargs["cov_struct"] = cov_struct
    day_range = f"{daily[DAYS_COL].min()}-{daily[DAYS_COL].max()}"
    print(
        f"Fitting generalized linear model on daily datasets. Days {day_range}\n"
        f"Model: {response_var} ~ days_from_surgery * diet."
        + (f" Covariance structure ({cov_struct})" if cov_struct else "")
    )

    if "1" in daily.diet.unique():
        daily.diet = daily.diet.map({"1": "Control", "2": "Fortasyn"})

    sig_pvalues = []

    if not "Overall" in daily.period.unique():
        # Sum Light and Dark to 'Overall' column
        daily_overall = daily.groupby(
            ["days_from_surgery", "cage"], as_index=False
        ).agg({response_var: "sum", "diet": "first"})
    else:
        daily_overall = daily[daily.period == "Overall"]

    for period in [DARK, LIGHT, "Overall"]:
        subset = daily[daily.period == period] if period != "Overall" else daily_overall

        md = smf.gee(
            f" {response_var} ~ days_from_surgery * diet",
            data=subset,
            groups="cage",
            **kwargs,
        )
        mdf = md.fit()
        if verbose:
            print(period + " GLM RESULTS:\n", mdf.summary())
        pvalues = mdf.pvalues
        for factor, pvalue in pvalues.items():
            if pvalue < 0.05 and factor != "Intercept":
                print(
                    f"{factor} effect on {response_var} ({period}) - p-value: {pvalue:.4f}"
                )
                sig_pvalues.append((pvalue, factor, period))
        if verbose:
            print(f"END {period} GLM RESULTS\n")

    for diet in ["Control", "Fortasyn"]:
        for period in [DARK, LIGHT, "Overall"]:
            subset = (
                daily.loc[(daily.period == period) & (daily.diet.str.contains(diet))]
                if period != "Overall"
                else daily_overall
            )
            if subset.empty:
                import ipdb

                ipdb.set_trace()
            md = smf.gee(
                f" {response_var} ~ days_from_surgery",
                data=subset,
                groups="cage",
                **kwargs,
            )
            mdf = md.fit()
            if verbose:
                print(f"{diet}-only, {period} GLM RESULTS:\n", mdf.summary())
            pvalues = mdf.pvalues
            for factor, pvalue in pvalues.items():
                if pvalue < 0.05 and factor != "Intercept":
                    print(
                        f"{factor} effect on {response_var} ({period}) - p-value: {pvalue:.4f}"
                    )
                    sig_pvalues.append((pvalue, factor, period))
    if not sig_pvalues:
        print("No significant effect found.\n")
    return sig_pvalues


class DVCExperiment:
    """Mouse tracking datasets analysis.
    experiment_name
    centriods_dir
    meta_filepath
    cage_xmax
    cage_ymax

    Args:

    Returns:

    """

    def __init__(
        self,
        experiment_name,
        centroids_dir: Optional[str] = None,
        # meta_filepath="/Users/justinshenk/neurodata/datasets/Stroke_olive_oil/DVC cageids HT Maximilian Wiesmann updated.xlsx",
        meta_filepath="/Users/justinshenk/neurodata/datasets/mouse_ref.csv",
        cage_xmax=0.058 * 2,
        cage_ymax=0.125 * 2,
    ):
        # TODO: Fix in prod version
        self.basedir = "/Users/justinshenk/neurodata/"
        self._cpu_count = mp.cpu_count()
        self.centroids_dir = centroids_dir
        self.mouse_lookup = self.load_meta(meta_filepath)
        self.cage_xmax = cage_xmax
        self.cage_ymax = cage_ymax
        self.experiment_name = experiment_name
        self._init()
        self.outdir = os.path.join(
            self.basedir, "output", self._str2filename(experiment_name)
        )
        self.cages = self.get_cages()

    def _init(self):
        """ """
        if self.centroids_dir:
            search_path = glob.glob(os.path.join(self.centroids_dir, "*"))
            self.centroid_files = sorted(
                [
                    x.split("/")[-1]
                    for x in search_path
                    if "csv" in x and "filelist" not in x
                ]
            )
        else:
            self.centroid_files = None

        # plt.rc("font", family="serif")
        style.use("ggplot")
        sns.set_context("paper")

    @staticmethod
    def _str2filename(string):
        """

        Args:
          string: 

        Returns:

        """
        filename = string.replace(" ", "_")
        # TODO: Implement filename security
        filename = filename.replace("/", "")
        return filename

    def get_weekly_activity(self):
        """ """
        activity = self.get_daily_activity(days=33)
        weekly_list = []

        for week in range(-3, 5):
            for group in activity["Group+Diet"].unique():
                for period in [LIGHT, DARK]:
                    df = (
                        activity[
                            (activity[DAYS_COL] >= week * 7 + 1)  # ...-6, 1, 8, 15...
                            & (
                                activity[DAYS_COL] < (week + 1) * 7 + 1
                            )  # ...1, 8, 15, 21...
                            & (activity["Group+Diet"] == group)
                            & (activity.period == period)
                        ]
                        .groupby(["Cage"])
                        .Activity.mean()
                        .to_frame()
                    )
                    df["Group+Diet"] = group
                    df["Week"] = week
                    df["Period"] = period
                    # df['Cohort'] = [get_cohort(x) for x in df.index]
                    weekly_list.append(df)
        weekly = pd.concat(weekly_list)
        return weekly

    def ttests(
        self,
        daily: pd.DataFrame,
        days: List[int],
        day_pairs: List[tuple],
        response_var: str,
    ) -> list:
        """Returns pvalues for intra and interday variances."""
        pvalues = []

        # Paired Student's t-Test, between days for each group
        for day1, day2 in day_pairs:
            diets = sorted(daily.diet.unique())
            for period in daily.period.unique():
                day1_control = daily.loc[
                    (daily["diet"] == diets[0])
                    & (daily[DAYS_COL] == day1)
                    & (daily.period == period),
                    response_var,
                ]
                day2_control = daily.loc[
                    (daily["diet"] == diets[0])
                    & (daily[DAYS_COL] == day2)
                    & (daily.period == period),
                    response_var,
                ]
                day1_fortasyn = daily.loc[
                    (daily["diet"] == diets[1])
                    & (daily[DAYS_COL] == day1)
                    & (daily.period == period),
                    response_var,
                ]
                day2_fortasyn = daily.loc[
                    (daily["diet"] == diets[1])
                    & (daily[DAYS_COL] == day2)
                    & (daily.period == period),
                    response_var,
                ]

                if day1_control.shape == day2_control.shape:
                    try:
                        control_test = (
                            ttest_rel(day1_control, day2_control).pvalue,
                            (day1, day2, "Control"),
                            period,
                        )
                    except AttributeError:
                        print(f"Error: Empty datasets, skipping Days: {day1}, {day2}")
                        continue
                    pvalues.append(control_test)
                else:
                    print(
                        f"Control ({period}) - Skipping stats - Day {day1} shape is {len(day1_control)} and day {day2} shape is {len(day2_control)}"
                    )
                if day1_fortasyn.shape == day2_fortasyn.shape:
                    try:
                        fortasyn_test = (
                            ttest_rel(day1_fortasyn, day2_fortasyn).pvalue,
                            (day1, day2, "Fortasyn"),
                            period,
                        )
                    except AttributeError:
                        print("Error: Empty datasets, skipping Days: {day1}, {day2}")
                        continue
                    pvalues.append(fortasyn_test)
                else:
                    print(
                        f"Fortasyn ({period}) - Skipping stats - Day {day1} shape is {len(day1_fortasyn)} and day {day2} shape is {len(day2_fortasyn)}"
                    )

        # Get intraday significance
        for day in daily[DAYS_COL].unique():
            diets = sorted(daily.diet.unique())

            # Student's t-Test, between groups for given day
            for period in daily.period.unique():
                control = daily.loc[
                    (daily["diet"] == diets[0])
                    & (daily[DAYS_COL] == day)
                    & (daily.period == period),
                    response_var,
                ]
                fortasyn = daily.loc[
                    (daily["diet"] == diets[1])
                    & (daily[DAYS_COL] == day)
                    & (daily.period == period),
                    response_var,
                ]
                pvalues.append((ttest_ind(control, fortasyn).pvalue, day, period))
        return pvalues

    def plot_daily_distance(
        self,
        daily: Optional[pd.DataFrame] = None,
        days: list = [3],
        metric: str = "distance",
        day_pairs: list = [],
        **kwargs,
    ):
        """Plot daily distance.

        Args:

           daily:   dataframe with daily datasets
           days:    cutoff for plots, eg [3,7] plots with maximum 3 days and 7 days.

        """
        if daily is None:
            daily = self.get_distance(time="D")

        daily = daily.copy()

        cages = sorted(daily.cage.unique())
        if not "days_from_surgery" in daily:  # days_from_stroke provided
            raise Exception("`days_from_surgery` not in dataframe")

        diet_names = ["Control", "Fortasyn"]
        if sorted(daily.diet.unique().tolist()) != sorted(diet_names):
            for ind, diet in enumerate(sorted(daily.diet.unique())):
                daily["diet"] = daily.diet.replace(diet, diet_names[ind])
                print(f"Renamed diet column value {diet} to {diet_names[ind]}")

        # Save original name for diet column
        daily_diet_orig = daily.diet
        pvalues = self.ttests(daily, days, day_pairs, response_var="displacement")

        # Calculate and mark significant difference
        for max_days in days:
            fig, ax = plt.subplots(figsize=(14, 8))
            plt.title(f"Average daily {metric} {max_days} days after stroke")

            if kwargs.get("t_test"):
                self.plot_t_test(daily, pvalues, max_days, response_var="displacement")

            daily.diet = daily.diet.str.cat(" - " + daily.period)
            ax = sns.pointplot(
                x="days_from_surgery",
                y="displacement",
                data=daily.loc[(daily["days_from_surgery"] <= max_days)],
                markers=["d", "s", "^", "x"],
                linestyles=["--", "--", "-", "-"],
                palette=["b", "b", "r", "r"],
                hue="diet",
                hue_order=sorted(daily.diet.unique()),
                dodge=True,
                conf_lw=0.8,
                ci=68,  # SEM
            )

            plt.legend(loc="upper left")

            # Revert to original column
            daily.diet = daily_diet_orig

            plt.xlabel("Days from Surgery")
            if metric == "velocity":
                ylabel = "Velocity (cm/s)"
            elif metric == "distance":
                ylabel = "Distance Travelled (cm)"
            plt.ylabel(ylabel)

            plt.show()

    def plot_weekly(self, weekly: TrajaDataFrame, groups):
        """

        Args:
          weekly:
          groups: 

        Returns:

        """
        for group in groups:
            fig, ax = plt.subplots(figsize=(4, 3))
            for period in [LIGHT, DARK]:
                sns.pointplot(
                    x="Week",
                    y="Activity",
                    hue="Cohort",
                    data=weekly[
                        (weekly["Group+Diet"] == group) & (weekly["Period"] == period)
                    ]
                    .groupby("Activity")
                    .mean()
                    .reset_index(),
                    ci=68,
                )
            plt.title(group)
            handles, labels = ax.get_legend_handles_labels()
            # sort both labels and handles by labels
            labels, handles = zip(
                *sorted(zip(labels[:2], handles[:2]), key=lambda t: t[0])
            )
            ax.legend(handles, labels)
            # plt.tight_layout()
        plt.show()

    def get_presurgery_average_weekly_activity(self):
        """Average pre-stroke weeks into one point."""
        pre_average_weekly_act = os.path.join(self.outdir, "pre_average_weekly_act.csv")
        if not os.path.exists(pre_average_weekly_act):
            weekly = self.get_weekly_activity()
            for period in [LIGHT, DARK]:
                for cage in self.get_cages():
                    mean = weekly[
                        (weekly.index == cage)
                        & (weekly.Week < 0)
                        & (weekly.Period == period)
                    ].Activity.mean()
                    weekly.loc[
                        (weekly.index == cage)
                        & (weekly.Week < 0)
                        & (weekly.Period == period),
                        "Activity",
                    ] = mean
        else:
            weekly = self.read_csv(pre_average_weekly_act)
            return weekly

    def norm_weekly_activity(self, weekly):
        """

        Args:
          weekly: 

        Returns:

        """
        # Normalize activity
        weekly["Normed_Activity"] = 0
        for period in [LIGHT, DARK]:
            for cage in self.get_cages():
                df_night = weekly[
                    (weekly["Week"] >= -1)
                    & (weekly.index == cage)
                    & (weekly.Period == DARK)
                ]
                df = weekly[
                    (weekly["Week"] >= -1)
                    & (weekly.index == cage)
                    & (weekly.Period == period)
                ]
                assert df.Week.is_monotonic_increasing == True, "Not monotonic"
                normed = [x / df_night.Activity.values[0] for x in df.Activity.values]
                weekly.loc[
                    (weekly.index == cage)
                    & (weekly.Period == period)
                    & (weekly.Week >= -1),
                    "Normed_Activity",
                ] = normed
        return weekly

    def _stylize_axes(self, ax):
        """

        Args:
          ax: 

        Returns:

        """
        ax.spines["top"].set_visible(False)
        ax.spines["right"].set_visible(False)

        ax.xaxis.set_tick_params(top="off", direction="out", width=1)
        ax.yaxis.set_tick_params(right="off", direction="out", width=1)

    def _shift_xtick_labels(self, xtick_labels, first_index=None):
        """

        Args:
          xtick_labels: param first_index:  (Default value = None)
          first_index:  (Default value = None)

        Returns:

        """
        for idx, x in enumerate(xtick_labels):
            label = x.get_text()
            xtick_labels[idx].set_text(str(int(label) + 1))
            if first_index is not None:
                xtick_labels[0] = first_index
        return xtick_labels

    def _norm_daily_activity(self, activity):
        """

        Args:
          activity: 

        Returns:

        """
        norm_daily_activity_csv = os.path.join(self.outdir, "norm_daily_activity.csv")

        if not os.path.exists(norm_daily_activity_csv):
            activity["Normed_Activity"] = 0
            for period in [LIGHT, DARK]:
                for cage in self.get_cages():
                    # Get prestroke
                    prestroke_night_average = activity[
                        (activity[DAYS_COL] <= -1)
                        & (activity.cage == cage)
                        & (activity.Period == DARK)
                    ].Activity.mean()
                    df = activity[
                        (activity[DAYS_COL] >= -1)
                        & (activity.cage == cage)
                        & (activity.Period == period)
                    ]
                    assert df[DAYS_COL].is_monotonic_increasing == True, "Not monotonic"
                    mean = activity[
                        (activity[DAYS_COL] <= -1)
                        & (activity.cage == cage)
                        & (activity.Period == period)
                    ].Activity.mean()
                    df.loc[
                        (df.Cage == cage)
                        & (df.Period == period)
                        & (df[DAYS_COL] == -1),
                        "Activity",
                    ] = mean
                    normed = [x / prestroke_night_average for x in df.Activity.values]
                    activity.loc[
                        (activity.Cage == cage)
                        & (activity.Period == period)
                        & (activity[DAYS_COL] >= -1),
                        "Normed_Activity",
                    ] = normed
            activity.to_csv(norm_daily_activity_csv)
        else:
            activity = pd.read_csv(norm_daily_activity_csv)
        return activity

    def plot_daily_normed_activity(self):
        """ """
        activity = self.get_daily_activity(days=33)
        normed_activity = self._norm_daily_activity(activity)
        ...

    def plot_weekly_normed_activity(self, presurgery_average=True):
        """Plot weekly normed activity. Optionally, average presurgery points.

        Args:
          presurgery_average: Default value = True)

        Returns:

        """
        if presurgery_average:
            weekly = self.get_presurgery_average_weekly_activity()
            # for cohort in [2,4]:
            fig, ax = plt.subplots(figsize=(6.25, 3.8))
            hue_order = weekly["Group+Diet"].unique()
            group_cnt = len(hue_order)
            for period in [LIGHT, DARK]:
                linestyles = (
                    ["--"] * group_cnt if period is LIGHT else ["-"] * group_cnt
                )
                sns.pointplot(
                    x="Week",
                    y="Normed_Activity",
                    hue="Group+Diet",
                    data=weekly[(weekly.Week >= -1) & (weekly.Period == period)],
                    #                                                                               (weekly.Cohort==cohort)],
                    palette=["k", "gray", "C0", "C1"][:group_cnt],
                    linestyles=linestyles,
                    # hue_order=['Sham - Control', 'Sham - HT', 'Stroke - Control', 'Stroke - HT'],
                    hue_order=hue_order,
                    markers=["d", "s", "^", "x"][
                        :group_cnt
                    ],  # TODO: Generalize for larger sets
                    dodge=True,
                    ci=68,
                )
            ax.set_xlabel("Weeks from Surgery")
            handles, labels = ax.get_legend_handles_labels()
            # sort both labels and handles by labels
            labels, handles = zip(
                *sorted(zip(labels[:4], handles[:4]), key=lambda t: t[0])
            )
            ax.legend(handles, labels)
            self._stylize_axes(ax)
            fig.set_facecolor("white")
            xtick_labels = ax.get_xticklabels()
            xtick_labels = self._shift_xtick_labels(xtick_labels, "Pre-surgery")

            plt.ylabel("Normalized Activity")
            ax.set_xticklabels(xtick_labels)
            plt.title("Normalized Activity")
            plt.show()

    def load_meta(self, meta_filepath):
        """

        Args:
          meta_filepath: 

        Returns:

        """
        if "xlsx" in meta_filepath:
            mouse_data = pd.read_excel(meta_filepath)[
                ["position", "Diet", "Sham_or_Stroke", "Stroke"]
            ]
            mouse_data["position"] = mouse_data["position"].apply(
                lambda x: x[1] + x[0].zfill(2)
            )
            return mouse_data.set_index("position").to_dict("index")
        elif "csv" in meta_filepath:
            mouse_ref = pd.read_csv("~/neurodata/datasets/mouse_ref.csv", sep="\t")
            mouse_ref = (
                mouse_ref.drop_duplicates().set_index("cage").drop(columns="mouse_nr")
            )

            return mouse_ref.to_dict()["diet"]

    def get_diet(self, cage):
        """

        Args:
          cage: 

        Returns:

        """

        diet = self.mouse_lookup[cage]
        if isinstance(diet, dict):
            return diet["Diet"]
        else:
            return diet

    def get_group(self, cage):
        """

        Args:
          cage: 

        Returns:

        """
        return self.mouse_lookup[cage]["Sham_or_Stroke"]

    def get_stroke(self, cage):
        """

        Args:
          cage: 

        Returns:

        """
        return self.mouse_lookup[cage]["Stroke"]

    def get_group_and_diet(self, cage):
        """

        Args:
          cage: 

        Returns:

        """
        diet = self.get_diet(cage)
        surgery = self.get_group(cage)
        return f"{'Sham' if surgery is 1 else 'Stroke'} - {'Control' if diet is 1 else 'HT'}"

    def get_cohort(self, cage):
        """

        Args:
          cage: 

        Returns:

        """
        # TODO: Generalize
        return self.mouse_lookup[cage]["Stroke"].month

    def get_cages(self):
        """ """
        return [x for x in self.mouse_lookup.keys()]

    def get_turn_ratios(self, file, angle_thresh, distance_thresh):
        """

        Args:
          file: param angle_thresh:
          distance_thresh: 
          angle_thresh: 

        Returns:

        """
        ratios = []
        cage = file.split("/")[-1].split("_")[0]
        # Get x,y coordinates from centroids
        date_parser = lambda x: pd.datetime.strptime(x, "%Y-%m-%d %H:%M:%S:%f")
        df = traja.read_file(file, index_col="time_stamps_vec")[["x", "y"]]
        # df.x = df.x.round(7)
        # df.y = df.y.round(7)
        df.traja.calc_distance()  # adds 'distance' column
        # TODO: Replace with generic intervention method name and lookup logic
        surgery_date = self.get_stroke(cage)
        df["Days_from_surgery"] = (df.index - surgery_date).days

        df.traja.calc_turn_angle()  # adds 'turn_angle' column
        #     df['turn_angle'].where((df['distance']>1e-3) & ((df.turn_angle > -15) & (df.turn_angle < 15))).hist(bins=30)
        #     df['turn_bias'] = df['turn_angle'] / .25 # 0.25s
        # Only look at distances over .01 meters, resample to minute intervals
        distance_mask = df["distance"] > (distance_thresh)
        angle_mask = ((df.turn_angle > angle_thresh) & (df.turn_angle < 90)) | (
            (df.turn_angle < -angle_thresh) & (df.turn_angle > -90)
        )

        day_mask = (df.index.hour >= 7) & (df.index.hour < 19)
        day_mean = df.loc[distance_mask & angle_mask & day_mask, "turn_angle"].dropna()
        night_mean = df.loc[
            distance_mask & angle_mask & ~day_mask, "turn_angle"
        ].dropna()
        right_turns_day = day_mean[day_mean > 0].shape[0]
        left_turns_day = day_mean[day_mean < 0].shape[0]
        right_turns_night = night_mean[night_mean > 0].shape[0]
        left_turns_night = night_mean[night_mean < 0].shape[0]
        ratios.append((df.Days_from_surgery[0], right_turns_day, left_turns_day, False))
        ratios.append(
            (df.Days_from_surgery[0], right_turns_night, left_turns_night, True)
        )

        ratios = [
            (day, right, left, period)
            for day, right, left, period in ratios
            if (left + right) > 0
        ]  # fix div by 0 error
        return ratios
        #     days = [day for day, _, _, nighttime in ratios if nighttime]

        #     laterality = [right_turns/(left_turns+right_turns) for day, right_turns, left_turns, nighttime in ratios if nighttime]
        #     fig, ax = plt.subplots()
        #     ax.plot(days, laterality, label='Laterality')
        #     ax.set(title=f"{cage} laterality index (right/right+left)\nDistance threshold: 0.25 cm\nAngle threshold: {thresh}\nRight turn is > 0.5\n{get_diet(cage)}",
        #           xlabel="Days from surgery",
        #           ylabel="Laterality index")
        #     ax.legend()
        #     ax.set_ylim((0,1.0))
        #     ax2 = ax.twinx()
        #     ax2.plot(days, [right+left for _, right, left, nighttime in ratios if nighttime],color='C1', label='Number of turns')
        #     ax2.set_ylabel('Number of turns')
        #     ax2.legend()
        #     plt.show()

    def calculate_turns(self, angle_thresh=30, distance_thresh=0.0025):
        """

        Args:
          angle_thresh: Default value = 30)
          distance_thresh: Default value = 0.0025)

        Returns:

        """
        ratio_dict = {}
        for cage in self.get_cages():
            ratio_dict[cage] = []

            with mp.Pool(processes=self._cpu_count) as p:
                args = [
                    (file, angle_thresh, distance_thresh)
                    for file in self.centroid_files
                    if cage in file
                ]
                ratios = p.starmap(self.get_ratios, args)
                ratio_dict[cage].append(ratios)
                logging.info(f"Processed {cage}")

        turn_ratio_csv = os.path.join(
            self.outdir,
            f"ratios_angle-{angle_thresh}_distance-{distance_thresh}_period_turnangle.npy",
        )
        np.save(turn_ratio_csv, ratio_dict)
        logging.info(f"Saved to {turn_ratio_csv}")
        return ratio_dict

    def get_coords(self, cage):
        """

        Args:
          cage: 

        Returns:

        """
        path = os.path.join(self.outdir, "centroids", cage)
        df = traja.read_file(path)
        return df

    @staticmethod
    def _get_distance(file, dates, diet):
        try:
            time = "D"  # daily
            basename = os.path.basename(file)
            cage = basename.split("_")[0]
            date = "".join(basename.split("_raw_")[0].split("_")[1:])
            days_from_surgery = dates.index(date)
            df = traja.read_file(
                file, index_col="time_stamps_vec", parse_dates=["time_stamps_vec"]
            )
            displacement = df.traja.calc_displacement()
            displacement.dropna(inplace=True)
            # Resample
            displacement = displacement.to_frame()
            displacement["cage"] = cage
            displacement["diet"] = diet
            displacement["days_from_surgery"] = days_from_surgery

            # Day
            day = (
                displacement.between_time("7:00", "19:00")
                .resample(time)
                .agg({"displacement": "sum"})
            )
            day["cage"] = cage
            day["diet"] = diet
            day["days_from_surgery"] = days_from_surgery
            day["period"] = LIGHT

            # Night
            night = (
                displacement.between_time("19:00", "7:00", include_end=False)
                .resample(time)
                .agg({"displacement": "sum"})
            )
            night["cage"] = cage
            night["diet"] = diet
            night["days_from_surgery"] = days_from_surgery
            night["period"] = DARK

            # Overall
            overall = displacement.resample(time).agg({"displacement": "sum"})
            overall["cage"] = cage
            overall["diet"] = diet
            overall["days_from_surgery"] = days_from_surgery
            overall["period"] = "Overall"
            return [day, night, overall]
        except Exception as e:
            print(e, file)

    def get_distance(self, time="D"):
        """Get distance travelled resampled by time.

        Args:
            time (str): 'D' is day, 'H' hour, 'T' minutes, 'S' seconds, and so on.

        Returns:

        """
        df_list = []

        time = time.lower()
        if time is "d" and hasattr(self, "daily_distance"):
            return self.daily_distance
        elif time == "d" and not hasattr(self, "daily_distance"):
            print(f"Calculating daily distance for {self.cages}")
            self.distance_files_processed = 0
            for cage in sorted(self.cages):
                cage_files = sorted([x for x in self.centroid_files if cage in x])
                dates = [
                    "".join(x.split("_raw_")[0].split("_")[1:]) for x in cage_files
                ]
                pool = mp.Pool(processes=self._cpu_count - 1)
                args = [
                    (
                        os.path.join(self.centroids_dir, file),
                        dates,
                        self.mouse_lookup[cage],
                    )
                    for file in cage_files
                ]
                results = pool.starmap(self._get_distance, iterable=args)
                pool.close()
                pool.join()

                if isinstance(results, list):
                    for dfs in results:
                        if isinstance(dfs, list):
                            try:
                                df_list.extend(dfs)
                            except TypeError:
                                pass
                            except Exception as e:
                                print(e)
                    print(
                        f"[{len(df_list)//3}/{len(self.centroid_files)}] - Calculated distance for {cage}"
                    )
                else:
                    print(f"Results are None for {cage}")

        try:
            daily_distance = pd.concat(df_list, ignore_index=True)
        except Exception as e:
            print(e)
            return None
        self.daily_distance = daily_distance
        return self.daily_distance

    def plot_position_heatmap(self, cage, bins=20):
        """

        Args:
          cage: param bins:  (Default value = 20)
          bins:  (Default value = 20)

        Returns:

        """
        from numpy import unravel_index

        # TODO: Generate from y in +-0.12, x in +-0.058
        try:
            x0, x1 = self._trj.xlim
            y0, y1 = self._trj.ylim
        except:
            raise NotImplementedError("Not yet implemented automated heatmap binning")
        x_edges = np.linspace(x0, x1, num=bins)
        y_edges = np.linspace(y0, y1, num=bins)

        trj = self.get_coords(cage)
        x, y = zip(*trj[["x", "y"]].values)
        # TODO: Remove redundant histogram calculation
        H, x_edges, y_edges = np.histogram2d(x, y, bins=(x_edges, y_edges))
        cmax = H.flatten().argsort()[-2]  # Peak point is too hot, bug?

        fig, ax = plt.subplots()
        hist, x_edges, y_edges, image = ax.hist2d(
            np.array(y),
            np.array(x),
            bins=[
                np.linspace(trj.y.min(), trj.y.max(), 50),
                np.linspace(trj.x.min(), trj.x.max(), 50),
            ],
            cmax=cmax,
        )
        ax.colorbar()
        ax.set_aspect("equal")
        plt.show()
        # peak_index = unravel_index(hist.argmax(),hist.shape)

    def get_activity_files(self):
        """ """
        activity_dir = os.path.join(
            self.basedir, "datasets", self.experiment_name, "dvc_activation", "*"
        )
        activity_files = glob.glob(activity_dir)
        assert activity_files, "No activity files"
        return activity_files

    def aggregate_files(self):
        """Aggregate cage files into csvs"""
        os.makedirs(os.path.join(self.outdir, "centroids"), exist_ok=True)
        for cage in self.centroid_files:
            logging.info(f"Processing {cage}")
            # Check for aggregated cage file (eg, 'A04.csv')
            cage_path = os.path.join(self.outdir, "centroids", f"{cage}.csv")
            if os.path.exists(cage_path):
                continue
            # Otherwise, generate one
            search_path = os.path.join(self.centroids_dir, cage, "*.csv")
            files = glob.glob(search_path)

            days = []
            for file in files:
                _df = self.read_csv(file)
                _df.columns = [x.strip() for x in _df.columns]
                days.append(_df)
            df = pd.concat(days).sort_index()
            #     for col in ['x','y','distance']:
            #         df.applymap(lambda x: x.str.strip() if isinstance(x,str) else x)
            #         df[col] = pd.to_numeric(df[col],errors='coerce')
            cage_path = os.path.join(self.outdir, "centroids", f"{cage}.csv")
            df.to_csv(cage_path)
            logging.info(f"saved to {cage_path}")
        # activity_df = self.read_csv('datasets/Stroke_olive_oil/dvc_activation/A04.csv', index_col='time_stamp_start')
        return

    def _get_ratio_dict(self, angle=30, distance=0.0025):
        """

        Args:
          angle: Default value = 30)
          distance: Default value = 0.0025)

        Returns:

        """
        npy_path = os.path.join(
            self.outdir,
            f"ratios_angle-{angle}_distance-{distance}_period_turnangle.npy",
        )
        r = np.load(npy_path)
        ratio_dict = r.item(0)
        return ratio_dict

    def get_cage_laterality(self, cage):
        """

        Args:
          cage: 

        Returns:

        """
        ratio_dict = self._get_ratio_dict()
        ratios = ratio_dict[cage]

        df = pd.DataFrame(ratios).stack().reset_index()
        df = df.drop(df.columns[[0, 1]], axis=1)

        df[["days", "right", "left", "Night"]] = pd.DataFrame(
            df[0].tolist(), index=df.index
        )
        df.drop(df.columns[[0]], axis=1, inplace=True)
        # ratios = [x for x in ratios if len(x) == 2 and (x[1] + x[2] > 0)]
        # days = [day for day, _, _, nighttime in ratios if nighttime]
        df["Laterality"] = df.right / (df.right + df.left)
        df["Turns"] = df.right + df.left

        fig, ax = plt.subplots()

        night = df.query("Night == True")
        ax = night.plot(x="days", y="Laterality", label="Laterality", ax=ax)
        thresh = 20
        ax.set(
            title=f"{cage} laterality index (right/right+left)\nDistance threshold: 0.25 cm\nAngle threshold: {thresh}\n"
            f"Right turn is > 0.5\n{self.get_diet(cage)}",
            xlabel="Days from surgery",
            ylabel="Laterality index",
        )
        ax.legend()
        ax.set_ylim((0, 1.0))
        ax2 = ax.twinx()
        ax2 = night.plot(
            x="days", y="Turns", label="Number of turns", ax=ax2, color="C1"
        )

        ax2.set_ylabel("Number of turns")
        ax2.legend()
        plt.show()

    def get_minute_activity_files(
        self, excel_path="datasets/DVC_Activity_20180702.xlsx", day_limit=37
    ):
        minute_activity_paths = []

        # Look for preexisting files
        for cage in self.cages:
            cage_filepath = os.path.join(
                self.basedir,
                "datasets",
                self.experiment_name,
                cage.replace(" ", "_") + ".csv",
            )
            if os.path.exists(cage_filepath):
                minute_activity_paths.append(cage_filepath)
            else:
                print(f"Skipping {cage} because {cage_filepath} does not exist.")

        # If found, return
        if len(minute_activity_paths):
            return minute_activity_paths

        # Create [cage].csv files with activity resampled to minutes
        if not os.path.exists(excel_path):
            raise FileNotFoundError(f"{excel_path} not found.")
        xl = pd.ExcelFile(excel_path)
        r = re.compile("[A-Z]\d")
        activity_sheets = list(filter(r.match, xl.sheet_names))

        for sheet_name in activity_sheets:
            df = xl.parse(sheet_name)
            df = df.iloc[1:day_limit]  # remove post-37-day datasets
            df = df[df.columns[1:-76]]  # remove non activity columns
            cage_filepath = os.path.join(
                "datasets", self.experiment_name, sheet_name.replace(" ", "_") + ".csv"
            )
            minute_activity_paths.append(cage_filepath)
            df.to_csv(cage_filepath, index=False)
            print(f"Saving to {cage_filepath}")

        return minute_activity_paths

    def plot_glm(self, daily, pvalues, max_days, period_group, response_var, **kwargs):
        # Plot asterisk for significance
        for idx, (pval, variable, period) in enumerate(pvalues):
            if period not in daily.period.unique():
                continue
            day = (daily[DAYS_COL].min(), max_days)
            if pval >= 0.05:
                continue
            if pval < 0.05:
                if DAYS_COL in variable:
                    day_text = f"Day {day[0]} - Day {day[1]} - {period}"
                elif response_var in variable:
                    day_text = (
                        f"Day {day[0]} - Day {day[1]} - {response_var}"
                        if isinstance(day, tuple)
                        else f"Day {day}"
                    )
                print(f"PLOT GLM: {day_text} ({period}): {pval:.3f}")

            # y_factor = (
            #     2
            #     if daily.loc[(daily.period == DARK), response_var].quantile(0.7) > 100
            #     else 1
            # )
            y_factor = 1

            y, h, col = (
                daily.loc[
                    (daily[DAYS_COL] <= max_days) & (daily.period == period),
                    response_var,
                ].quantile(0.75)
                + y_factor * idx,
                0.1,
                "k",
            )

            text = "*" if pval < 0.05 else ""
            text = "**" if pval < 0.01 else text
            text = "***" if pval < 0.001 else text

            day_range = range(daily[DAYS_COL].min(), max_days + 1)

            if isinstance(day, tuple):
                # Prevent plotting pvalues that won't fit on plot
                if day[1] > max_days:
                    continue

                # Annotate interday signfificance
                x1, x2 = (
                    np.searchsorted(day_range, day[0]),
                    np.searchsorted(day_range, day[1]),
                )

                diet = f"{period} p <= {pval:.3f}"
                plt.plot([x1, x1, x2, x2], [y, y + h, y + h, y], lw=1.5, c=col)
                plt.text(
                    (x1 + x2) * 0.5,
                    y + h,
                    text + diet,
                    ha="center",
                    va="bottom",
                    color=col,
                    fontsize=14,
                )
            else:
                x1 = np.searchsorted(day_range, day)
                plt.text(
                    x1,
                    y + h,
                    text + f" p <= {pval:.3f}",
                    ha="center",
                    va="bottom",
                    color=col,
                    fontsize=14,
                )

    def plot_daily_activity(
        self,
        daily: Optional[pd.DataFrame] = None,
        days: list = [3],
        metric: str = "activity",
        day_pairs: list = [],
        **kwargs,
    ):
        if daily is None:
            daily = self.get_daily_activity(days)

        days = days or [daily[DAYS_COL].max()]

        assert DAYS_COL in daily, f"{DAYS_COL} not in dataframe"

        # For replacing numeric diet encoding
        diet_names = ["Control", "Fortasyn"]
        if sorted(daily.diet.unique().tolist()) != sorted(diet_names):
            for ind, diet in enumerate(sorted(daily.diet.unique())):
                daily["diet"] = daily.diet.replace(diet, diet_names[ind])
                print(f"Renamed diet column value {diet} to {diet_names[ind]}")

        # # Statistical analysis
        # pvalues = self.ttests(daily, days, day_pairs, response_var="activity")

        for max_days in days:
            fig, ax = plt.subplots(figsize=(14, 8))
            plt.title(f"Average daily {metric} {max_days} days after stroke")

            # if kwargs.get("t_test"):
            #     self.plot_t_test(daily, pvalues, max_days, response_var="activity")

            daily.diet = daily.diet.str.cat(" - " + daily.period)

            if kwargs.get("glm"):
                pvalues = glm(daily, response_var="activity")

            for period_group in ["Overall", "Phases"]:
                if period_group == "Overall":
                    subset = daily.loc[
                        (daily[DAYS_COL] <= max_days)
                        & (daily.Diet.str.contains("Overall"))
                    ]
                    markers = ["d", "s"]
                    linestyles = ["-", "-"]
                    palette = ["b", "r"]
                elif period_group == "Phases":
                    subset = daily.loc[
                        (daily[DAYS_COL] <= max_days)
                        & (~daily.Diet.str.contains("Overall"))
                    ]
                    markers = ["d", "s", "^", "x"]
                    linestyles = ["-", "--", "-", "--"]
                    palette = ["b", "b", "r", "r"]

                # self.plot_glm(
                #     daily, pvalues, max_days, period_group, response_var="activity"
                # )

                ax = sns.pointplot(
                    x=DAYS_COL,
                    y="activity",
                    data=subset,
                    markers=markers,
                    linestyles=linestyles,
                    palette=palette,
                    hue="Diet",
                    hue_order=sorted(subset.Diet.unique()),
                    dodge=True,
                    conf_lw=0.8,
                    ci=68,  # SEM
                )

                plt.legend(loc="upper left")

                plt.xlabel("Days from Surgery")
                ylabel = "Activity (a.u.)"
                plt.ylabel(ylabel)

                plt.show()
        ...

    def get_daily_activity(self, days: int) -> pd.DataFrame:
        """Returns daily activity."""
        minute_activity_files = self.get_minute_activity_files()

        # initialize
        activity = pd.DataFrame(columns=["cage", "activity", "Diet", DAYS_COL])

        # get minute activity files
        sorted_cages = self.get_minute_activity_files()

        def csv_to_series(cage_file, fill_nan=True):
            cage = pd.read_csv(cage_file)
            if cage.shape[1] != 1439:
                print(
                    "Warning: Cage {cage} has {cage.shape[1]} columns instead of 1439"
                )
            # nan_count = cage.isna().sum().sum()  # about 1440 * 8 NaNs
            if fill_nan:
                cage.fillna(0, inplace=True)
            flat = cage.values.flatten()
            cage_name = cage_file.split("/")[-1].split(".")[0]
            diet = self.mouse_lookup[cage_name]
            return flat, cage_name, diet

        for f in sorted_cages:
            flat, cage_name, diet = csv_to_series(f)
            day_indices = np.arange(390, 1110)
            mask = np.ones(1440, np.bool)
            mask[day_indices] = 0
            night_indices = np.arange(1440)[mask]
            for day in range(days):
                group = "Control" if int(diet) is 1 else "Fortasyn"
                overall_data = np.nanmean(flat[np.arange(1440) + day * 1440])
                day_data = np.nanmean(flat[day_indices + day * 1440])
                night_data = np.nanmean(flat[night_indices + day * 1440])
                activity = activity.append(
                    {
                        "activity": day_data,
                        "cage": cage_name,
                        "period": LIGHT,
                        "diet": group,
                        "Diet": group + f" - {LIGHT}",
                        DAYS_COL: int(day + 1),
                    },
                    ignore_index=True,
                )
                activity = activity.append(
                    {
                        "activity": night_data,
                        "cage": cage_name,
                        "period": DARK,
                        "diet": group,
                        "Diet": group + f" - {DARK}",
                        DAYS_COL: int(day + 1),
                    },
                    ignore_index=True,
                )
                activity = activity.append(
                    {
                        "activity": overall_data,
                        "cage": cage_name,
                        "period": "Overall",
                        "diet": group,
                        "Diet": group + " - Overall",
                        DAYS_COL: int(day + 1),
                    },
                    ignore_index=True,
                )

        activity[DAYS_COL] = activity[DAYS_COL].astype(int)

        return activity

    # def get_daily_activity(self):
    #     """ """
    #     activity_csv = os.path.join(self.outdir, "daily_activity.csv")
    #     if not os.path.exists(activity_csv):
    #         print(f"Path {activity_csv} does not exist, creating dataframe")
    #         activity_list = []
    #         col_list = [f"e{i:02}" for i in range(1, 12 + 1)]  # electrode columns
    #
    #         # Iterate over minute activations
    #         search_path = os.path.join(
    #             self.basedir, "datasets", self.experiment_name, "dvc_activation", "*.csv"
    #         )
    #         minute_activity_files = sorted(glob.glob(search_path))
    #         for cage in minute_activity_files:
    #             cage_id = os.path.split(cage)[-1].split(".")[0]
    #             # TODO: Fix in final
    #             assert len(cage_id) == 3, logging.error(f"{cage_id} length != 3")
    #             # Read csv
    #             cage_df = pd.read_csv(
    #                 cage,
    #                 index_col="time_stamp_start",
    #                 date_parser=lambda x: pd.datetime.strptime(
    #                     x, "%Y-%m-%d %H:%M:%S:%f"
    #                 ),
    #             )
    #             # Make csv with columns for cage+activity+day+diet+surgery
    #             cage_df["Activity"] = cage_df[col_list].sum(axis=1)
    #             day = (
    #                 cage_df.traja.day.groupby(pd.Grouper(key="time", freq="D"))[
    #                     "Activity"
    #                 ]
    #                 .sum()
    #                 .to_frame()
    #             )
    #             day["Cage"] = cage_id
    #             day["Period"] = "LIGHT"
    #             day["Surgery"] = self.get_stroke(cage_id)
    #             day["Diet"] = self.get_diet(cage_id)
    #             day["Group"] = self.get_group(cage_id)
    #             day["Days"] = [int(x) for x in range(len(day.index))]
    #             activity_list.append(day)
    #
    #             night = (
    #                 cage_df.traja.night.groupby(pd.Grouper(key="time", freq="D"))[
    #                     "Activity"
    #                 ]
    #                 .sum()
    #                 .to_frame()
    #             )
    #             night["Cage"] = cage_id
    #             night["Period"] = "DARK"
    #             night["Surgery"] = self.get_stroke(cage_id)
    #             night["Diet"] = self.get_diet(cage_id)
    #             night["Group"] = self.get_group(cage_id)
    #             night["Days"] = [int(x) for x in range(len(night.index))]
    #             activity_list.append(night)
    #
    #         activity = pd.concat(activity_list)
    #         activity.to_csv(activity_csv)
    #     else:
    #         activity = traja.read_file(
    #             activity_csv,
    #             index_col="time_stamp_start",
    #             parse_dates=["Surgery", "time_stamp_start"],
    #             infer_datetime_format=True,
    #         )
    #     return activity

    def animate(self, trajectory, timesteps=None):
        """Animate trajectory over time with statistical information about turn angle, etc.

        Args:
          trajectory: param timesteps:  (Default value = None)
          timesteps:  (Default value = None)

        Returns:

        """
        if timesteps is not None:
            df = trajectory.iloc[:timesteps]
        else:
            df = trajectory
        ratios = {"left": 0, "right": 0}
        thresh = 30

        # Scale to centimeters (optional)
        df.x *= 100
        df.y *= 100
        if not "distance" in trajectory:
            df.traja.calc_distance()
        df.distance *= 100

        df.traja.calc_turn_angle()
        #     df['turn_angle'].where((df['distance']>1e-3) & ((df.turn_angle > -15) & (df.turn_angle < 15))).hist(bins=30)
        df["turn_bias"] = df["turn_angle"] / 0.25  # 0.25s
        # Only look at distances over .01 meters, resample to minute intervals
        distance_mask = df["distance"] > 1e-2
        angle_mask = (df.turn_bias > thresh) | (df.turn_bias < -thresh)

        # coords
        cage_y = 25.71
        cage_x = 13.7

        fig, axes = plt.subplots(
            2, 1, figsize=(8, 6), gridspec_kw={"height_ratios": [9, 1]}
        )

        def col_func(val, minval, maxval, startcolor, stopcolor):
            """Convert value in the range minval...maxval to a color in the range
                startcolor to stopcolor. The colors passed and the one returned are
                composed of a sequence of N component values (e.g. RGB).

            Args:
              val: param minval:
              maxval: param startcolor:
              stopcolor:
              minval:
              startcolor:

            Returns:

            """
            f = float(val - minval) / (maxval - minval)
            return tuple(f * (b - a) + a for (a, b) in zip(startcolor, stopcolor))

        RED, YELLOW, GREEN = (1, 0, 0), (1, 1, 0), (0, 1, 0)
        CYAN, BLUE, MAGENTA = (0, 1, 1), (0, 0, 1), (1, 0, 1)
        steps = 10
        minval, maxval = 0.0, 1.0
        incr = (maxval - minval) / steps

        ax = axes[0]

        for i in df.index:
            ax.cla()
            ax.set_aspect("equal")
            ax.set_ylim(-cage_y / 2, cage_y / 2)
            ax.set_xlim(-cage_x / 2, cage_x / 2)

            x, y = df.loc[i, ["x", "y"]].values
            turn_bias = df.loc[i, "turn_bias"]

            # Scale to 0-1
            laterality = turn_bias + 360
            laterality /= 360 * 2
            if laterality > 1:
                laterality = 1
            elif laterality < 0:
                laterality = 0

            color = col_func(laterality, minval, maxval, BLUE, RED)
            ax.plot(x, y, color=color, marker="o")
            ax.invert_yaxis()

            try:
                # Filter for 1 cm/s
                # distance = df.distance.loc[i-2:i+2].sum()
                distance = df.distance.loc[i]
            except:
                print(f"Skipping {i}")
                continue

            count_turn = (distance >= 0.25) & ((turn_bias > 20) | (turn_bias < -20))
            if count_turn:
                if turn_bias > 0:
                    ratios["right"] += 1
                elif turn_bias < 0:
                    ratios["left"] += 1

            distance_str = (
                rf"$\bf{distance:.2f}$" if distance >= 0.25 else f"{distance:.2f}"
            )

            total_turns = ratios["right"] + ratios["left"]
            if total_turns == 0:
                overall_laterality = 0.5
            else:
                overall_laterality = ratios["right"] / total_turns

            ax.set_title(
                f"frame {i} - distance (cm/0.25s): {distance_str}\n \
                x: {x:.2f}, y: {y:.2f}\n \
                Heading: {df.loc[i,'heading']:5.0f} degrees\n \
                Turn Angle: {df.loc[i,'turn_angle']:4.0f}\n \
                Turn Bias: {turn_bias:4.0f}\n \
                Current Laterality: {laterality:.2f}\n \
                Left: {ratios['left']:>3}, Right: {ratios['right']:>3}\n \
                Overall Laterality: {overall_laterality:.2f} \
                "
            )

            axes[1].cla()
            if turn_bias > 0:
                rect = patches.Rectangle(
                    (0, 0), laterality, 1, linewidth=1, edgecolor="k", facecolor="r"
                )
            elif turn_bias < 0:
                rect = patches.Rectangle(
                    (0, 0), laterality, 1, linewidth=1, edgecolor="k", facecolor="b"
                )
            else:
                rect = patches.Rectangle(
                    (0, 0), laterality, 1, linewidth=1, edgecolor="k", facecolor="gray"
                )
            # Add the patch to the Axes
            axes[1].add_patch(rect)
            fig.tight_layout()
            plt.pause(0.01)

    def plot_t_test(self, daily, pvalues, max_days, response_var):
        # Plot asterisk for significance
        for idx, (pval, day, period) in enumerate(pvalues):
            if pval >= 0.05:
                continue
            if pval < 0.05:
                day_text = (
                    f"Day {day[0]} - Day {day[1]} - {day[2]}"
                    if isinstance(day, tuple)
                    else f"Day {day}"
                )
                print(f"{day_text} ({period}): {pval:.3f}")

            y_factor = (
                6
                if daily.loc[(daily.period == DARK), response_var].quantile(0.7) > 100
                else 2
            )

            y, h, col = (
                daily.loc[
                    (daily[DAYS_COL] <= max_days) & (daily.period == period),
                    response_var,
                ].quantile(0.75)
                + y_factor * idx,
                1.5,
                "k",
            )

            text = "*" if pval < 0.05 else ""
            text = "**" if pval < 0.01 else text
            text = "***" if pval < 0.001 else text

            day_range = range(daily[DAYS_COL].min(), max_days + 1)

            if isinstance(day, tuple):
                # Prevent plotting pvalues that won't fit on plot
                if day[1] > max_days:
                    continue

                # Annotate interday signfificance
                x1, x2 = (
                    np.searchsorted(day_range, day[0]),
                    np.searchsorted(day_range, day[1]),
                )

                diet = f"{day[2]} {period} p <= {pval:.3f}"
                plt.plot([x1, x1, x2, x2], [y, y + h, y + h, y], lw=1.5, c=col)
                plt.text(
                    (x1 + x2) * 0.5,
                    y + h,
                    text + diet,
                    ha="center",
                    va="bottom",
                    color=col,
                    fontsize=14,
                )
            else:
                x1 = np.searchsorted(day_range, day)
                plt.text(
                    x1,
                    y + h,
                    text + f" p <= {pval:.3f}",
                    ha="center",
                    va="bottom",
                    color=col,
                    fontsize=14,
                )


def get_plot(file, axes, col=0):
    import traja

    df = traja.read_file(file, parse_dates=["time_stamps_vec"])
    for row, period in enumerate(["day", "night"]):
        if period is "day":
            period_df = df.traja.day()
        else:
            period_df = df.traja.night()
        filename = file.split("/")[-1].split(".")[0] + f"_{period}"
        traja.plotting.polar_bar(
            period_df,
            ax=axes[row, col],
            title=filename,
            save=filename + ".png",
            show=False,
        )


def debug():
    F4 = [
        "F6_2016-01-30_raw_centroid_positions.csv",
        "F4_2016-02-06_raw_centroid_positions.csv",
        "F4_2016-02-28_raw_centroid_positions.csv",
    ]
    F6 = [
        "F6_2016-01-30_raw_centroid_positions.csv",
        "F6_2016-02-28_raw_centroid_positions.csv",
    ]

    fig, axes = plt.subplots(2, 3, subplot_kw=dict(polar=True))

    for col, file in enumerate(F4):
        file = os.path.join("~/neurodata/datasets/raw_centroids_rev2", file)
        get_plot(file, axes, col)
    plt.show()


def pair(s: str):
    try:
        day_range = s.split("-")
        day_range = (int(day_range[0]), int(day_range[1]))
        if len(day_range) > 2:
            raise argparse.ArgumentTypeError(
                f"Day range length can only be 2, but it is {len(day_range)}"
            )
        return day_range
    except:
        raise argparse.ArgumentTypeError(
            "Day range must be hyphen separated, eg: --day-range=1-7"
        )


def pairs(s: str):
    try:
        day_pairs = s.split(",")
        day_pairs = [x.split("-") for x in day_pairs]
        day_pairs = [(int(x[0]), int(x[1])) for x in day_pairs]
        return day_pairs
    except:
        raise argparse.ArgumentTypeError(
            "Day pairs must be comma separated with hyphens for range, eg: --day-pairs=1-7,1-21"
        )


def days(s: str):
    try:
        days = [int(x) for x in s.split(",")]
        return days
    except:
        raise argparse.ArgumentTypeError(
            "Days must be comma separated, eg: --days=1,7,21"
        )


def main(args):
    if not args.activity:
        daily_filename = f"daily_distance_{args.experiment_name}.csv"
        if os.path.exists(daily_filename):
            daily = pd.read_csv(daily_filename, dtype={"diet": "category"})
            print(f"Loaded {daily_filename}")
            experiment = DVCExperiment(experiment_name=args.experiment_name)
        elif args.centroids_dir and args.centroids_dir:
            centroids_dir = os.path.abspath(args.centroids_dir)
            if not os.path.exists(centroids_dir):
                raise Exception(f"Error: {centroids_dir} does not exist.")
            experiment = DVCExperiment(
                experiment_name=args.experiment_name, centroids_dir=centroids_dir
            )
            print(f"Computing daily distance.")
            daily = experiment.get_distance()
            daily = daily.drop_duplicates(
                daily, subset=["diet", "period", "cage", "days_from_surgery"]
            )
        else:
            raise Exception(
                f"Must provide centroids_dir and/or experiment_name, {daily_filename} not found"
            )
        if args.day_range:
            d1, d2 = args.day_range
            daily = daily.loc[
                (daily["days_from_surgery"] >= d1) & (daily["days_from_surgery"] <= d2)
            ]
        if args.day_pairs or args.days:
            print("Plotting daily distance")
            experiment.plot_daily_distance(daily, **args.__dict__)

        # Statistical analysis
        if args.glm:
            glm(daily, response_var="displacement", verbose=args.verbose)
    else:
        # Activity-only
        experiment = DVCExperiment(experiment_name=args.experiment_name)
        activity = experiment.get_daily_activity(33)
        d1, d2 = args.day_range
        activity = activity[(activity[DAYS_COL] >= d1) & (activity[DAYS_COL] <= d2)]
        print("Plotting daily activity")
        experiment.plot_daily_activity(activity, **args.__dict__)

    # Statistical analysis
    if args.glm and args.activity:
        glm(activity, response_var="activity", verbose=args.verbose)


def parse_args():
    parser = argparse.ArgumentParser(description="Run DVC Experiment")
    parser.add_argument(
        "-c", "--centroids_dir", help="path to directory with centroids CSVs"
    )
    parser.add_argument(
        "-n",
        "--name",
        dest="experiment_name",
        type=str,
        help="name of experiment, eg, fortasyn",
    )
    parser.add_argument(
        "-d",
        "--days",
        help="number of days per plot, comma-separated, eg: 3,7,35",
        type=days,
        default=[],
    )
    parser.add_argument(
        "-p",
        "--day_pairs",
        help="day pairs for statistical comparison, comma-separated, eg: 1-7,1-21",
        type=pairs,
        default=[],
    )
    parser.add_argument(
        "-g",
        "--glm",
        help="Estimate with generalized linear model",
        action="store_true",
    )
    parser.add_argument(
        "-r", "--day-range", help="Restrict day range, eg: 1-7", type=pair
    )
    parser.add_argument(
        "-t", "--t-test", help="Plot t-Test results", action="store_true"
    )
    parser.add_argument(
        "-a", "--activity", help="Plot daily activity", action="store_true"
    )
    parser.add_argument(
        "-v", "--verbose", help="print all stats details", action="store_true"
    )
    args = parser.parse_args()

    max_days = 0
    if args.day_pairs:
        max_days = max(max(t) for t in args.day_pairs)
    if args.days:
        max_days = max(max_days, max(args.days))
    if args.day_range and args.days and args.day_range[-1] < max_days:
        raise Exception(
            f"Invalid options for day range: Day range {args.day_range}, Days {args.days}"
        )
    return args


if __name__ == "__main__":
    args = parse_args()
    main(args)



import math
from collections import OrderedDict
from typing import Callable, Optional, Union, Tuple

import numpy as np
import pandas as pd
from pandas.core.dtypes.common import (
    is_datetime_or_timedelta_dtype,
    is_datetime64_any_dtype,
    is_timedelta64_dtype,
)
from scipy import signal
from scipy.spatial.distance import directed_hausdorff, euclidean

import traja
from traja import TrajaDataFrame


__all__ = [
    "_bins_to_tuple",
    "_get_time_col",
    "_get_xylim",
    "_grid_coords1D",
    "_has_cols",
    "_rediscretize_points",
    "_resample_time",
    "angles",
    "calc_angle",
    "calc_derivatives",
    "calc_displacement",
    "calc_heading",
    "calc_turn_angle",
    "calculate_flow_angles",
    "cartesian_to_polar",
    "coords_to_flow",
    "distance_between",
    "distance",
    "euclidean",
    "expected_sq_displacement",
    "fill_in_traj",
    "from_xy",
    "generate",
    "get_derivatives",
    "grid_coordinates",
    "length",
    "polar_to_z",
    "rediscretize_points",
    "resample_time",
    "rotate",
    "smooth_sg",
    "speed_intervals",
    "step_lengths",
    "to_shapely",
    "traj_from_coords",
    "transition_matrix",
    "transitions",
]


def smooth_sg(trj: TrajaDataFrame, w: int = None, p: int = 3):
    """Returns ``DataFrame`` of trajectory after Savitzky-Golay filtering.

    Args:
      trj (:class:`~traja.frame.TrajaDataFrame`): Trajectory
      w (int): window size (Default value = None)
      p (int): polynomial order (Default value = 3)

    Returns:
      trj (:class:`~traja.frame.TrajaDataFrame`): Trajectory

    .. doctest::

        >> df = traja.generate()
        >> traja.smooth_sg(df, w=101).head()
                   x          y  time
        0 -11.194803  12.312742  0.00
        1 -10.236337  10.613720  0.02
        2  -9.309282   8.954952  0.04
        3  -8.412910   7.335925  0.06
        4  -7.546492   5.756128  0.08

    """
    if w is None:
        w = p + 3 - p % 2

    if w % 2 != 1:
        raise Exception(f"Invalid smoothing parameter w ({w}): n must be odd")
    _trj = trj.copy()
    _trj.x = signal.savgol_filter(_trj.x, window_length=w, polyorder=p, axis=0)
    _trj.y = signal.savgol_filter(_trj.y, window_length=w, polyorder=p, axis=0)
    _trj = fill_in_traj(_trj)
    return _trj


def angles(trj: TrajaDataFrame, lag: int = 1):
    """Returns angles w.r.t. x-axis."""
    dx = trj.x.diff(lag)
    distance = calc_displacement(trj, lag=lag)
    angles = np.rad2deg(np.arccos(np.abs(dx) / distance))
    # Correction for 360-degree angle range
    angles[angles >= 180] -= 360
    angles[angles < -180] += 360
    return angles


def apply_all(trj: TrajaDataFrame, method: Callable, id_col: str, **kwargs):
    """Applies method to all trajectories"""
    return trj.groupby(by=id_col).apply(method, **kwargs)


def step_lengths(trj: TrajaDataFrame):
    """Length of the steps of ``trj``.

    Args:
      trj (:class:`~traja.frame.TrajaDataFrame`): Trajectory

    """
    displacement = traja.trajectory.calc_displacement(trj)
    return displacement


def polar_to_z(r: float, theta: float) -> complex:
    """Converts polar coordinates ``r`` and ``theta`` to complex number ``z``.

    Args:
      r (float): step size
      theta (float): angle

    Returns:
        z (complex): complex number z

    """
    return r * np.exp(1j * theta)


def cartesian_to_polar(xy: np.ndarray) -> (float, float):
    """Convert :class:`numpy.ndarray` ``xy`` to polar coordinates ``r`` and ``theta``.

    Args:
      xy (:class:`numpy.ndarray`): x,y coordinates

    Returns:
        r, theta (tuple of float): step-length and angle

    """
    assert xy.ndim == 2, f"Dimensions are {xy.ndim}, expecting 2"
    x, y = np.split(xy, [-1], axis=1)
    x, y = np.squeeze(x), np.squeeze(y)
    r = np.sqrt(x * x + y * y)
    theta = np.arctan2(y, x)
    return r, theta


def distance(trj: TrajaDataFrame) -> float:
    """Calculates the distance from start to end of trajectory, also called net distance, displacement, or bee-line
    from start to finish.

    Args:
        trj (:class:`~traja.frame.TrajaDataFrame`): Trajectory

    Returns:
        distance (float)

    .. doctest::

        >> df = traja.generate()
        >> traja.distance(df)
        117.01507823153617

    """
    start = trj.iloc[0][["x", "y"]].values
    end = trj.iloc[-1][["x", "y"]].values
    return np.linalg.norm(end - start)


def length(trj: TrajaDataFrame) -> float:
    """Calculates the cumulative length of a trajectory.

    Args:
        trj (:class:`~traja.frame.TrajaDataFrame`): Trajectory

    Returns:
        length (float)

    .. doctest::

        >> df = traja.generate()
        >> traja.length(df)
        2001.142339606066

    """
    displacement = trj.traja.calc_displacement()
    return displacement.sum()


def expected_sq_displacement(
    trj: TrajaDataFrame, n: int = 0, eqn1: bool = True
) -> float:
    """Expected displacement.

    .. note::

        This method is experimental and needs testing.

    """
    sl = traja.step_lengths(trj)
    ta = traja.angles(trj)
    l = np.mean(sl)
    l2 = np.mean(sl ** 2)
    c = np.mean(np.cos(ta))
    s = np.mean(np.sin(ta))
    s2 = s ** 2

    if eqn1:
        # Eqn 1
        alpha = np.arctan2(s, c)
        gamma = ((1 - c) ** 2 - s2) * np.cos((n + 1) * alpha) - 2 * s * (
            1 - c
        ) * np.sin((n + 1) * alpha)
        esd = (
            n * l2
            + 2 * l ** 2 * ((c - c ** 2 - s2) * n - c) / ((1 - c) ** 2 + s2)
            + 2
            * l ** 2
            * ((2 * s2 + (c + s2) ** ((n + 1) / 2)) / ((1 - c) ** 2 + s2) ** 2)
            * gamma
        )
        return abs(esd)
    else:
        print("This method is experimental and requires testing.")
        # Eqn 2
        esd = n * l2 + 2 * l ** 2 * c / (1 - c) * (n - (1 - c ** n) / (1 - c))
        return esd


def traj_from_coords(
    track: Union[np.ndarray, pd.DataFrame],
    x_col=1,
    y_col=2,
    time_col: Optional[str] = None,
    fps: Union[float, int] = 4,
    spatial_units: str = "m",
    time_units: str = "s",
) -> TrajaDataFrame:
    """Create TrajaDataFrame from coordinates.

    Args:
        track:  N x 2 numpy array or pandas DataFrame with x and y columns
        x_col: column index or x column name
        y_col: column index or y column name
        time_col: name of time column
        fps: Frames per seconds
        spatial_units: default m, optional
        time_units: default s, optional

    Returns:
        trj: TrajaDataFrame

    .. doctest::

        >> xy = np.random.random((1000, 2))
        >> trj = traja.traj_from_coord(xy)
        >> assert trj.shape == (1000,4) # columns x, y, time, dt

    """
    if not isinstance(track, traja.TrajaDataFrame):
        if isinstance(track, np.ndarray) and track.shape[1] == 2:
            trj = traja.from_xy(track)
        elif isinstance(track, pd.DataFrame):
            trj = traja.TrajaDataFrame(track)
    else:
        trj = track
    trj.traja.spatial_units = spatial_units
    trj.traja.time_units = time_units

    def rename(col, name, trj):
        if isinstance(col, int):
            trj.rename(columns={col: name})
        else:
            if col not in trj:
                raise Exception(f"Missing column {col}")
            trj.rename(columns={col: name})
        return trj

    # Ensure column names are as expected
    trj = rename(x_col, "x", trj)
    trj = rename(y_col, "y", trj)
    if time_col is not None:
        trj = rename(time_col, "time", trj)

    # Allocate times if they aren't already known
    if "time" not in trj:
        if fps is None:
            raise Exception(
                (
                    "Cannot create a trajectory without times: either fps or a time column must be specified"
                )
            )
        # Assign times to each frame, starting at 0
        trj["time"] = pd.Series(np.arange(0, len(trj)) / fps)

    # Get displacement time for each coordinate, with the first point at time 0
    trj["dt"] = trj.time - trj.time.iloc[0]

    return trj


def distance_between(A: traja.TrajaDataFrame, B: traja.TrajaDataFrame, method="dtw"):
    """Returns distance between two trajectories.

    Args:
        A (:class:`~traja.frame.TrajaDataFrame`) : Trajectory 1
        B (:class:`~traja.frame.TrajaDataFrame`) : Trajectory 2
        method (str): ``dtw`` for dynamic time warping, ``hausdorff`` for Hausdorff

    Returns:
        distance (float): Distance

    """
    if method == "hausdorff":
        dist0 = directed_hausdorff(A, B)[0]
        dist1 = directed_hausdorff(B, A)[0]
        symmetric_dist = max(dist0, dist1)
        return symmetric_dist
    elif method == "dtw":
        try:
            from fastdtw import fastdtw
        except ImportError:
            raise ImportError(
                """            
            Missing optional dependency 'fastdtw'. Install fastdtw for dynamic time warping distance with pip install 
            fastdtw.
            """
            )
        distance, path = fastdtw(A, B, dist=euclidean)
        return distance


def to_shapely(trj):
    """Returns shapely object for area, bounds, etc. functions.

    Args:
        trj (:class:`~traja.frame.TrajaDataFrame`): Trajectory

    Returns:
      shapely.geometry.linestring.LineString -- Shapely shape.

    .. doctest::

        >>> df = traja.TrajaDataFrame({'x':[0,1,2],'y':[1,2,3]})
        >>> shape = traja.to_shapely(df)
        >>> shape.is_closed
        False

    """
    from shapely.geometry import shape

    coords = trj[["x", "y"]].values
    tracks_obj = {"type": "LineString", "coordinates": coords}
    tracks_shape = shape(tracks_obj)
    return tracks_shape


def transition_matrix(grid_indices1D: np.ndarray):
    """Returns ``np.ndarray`` of Markov transition probability matrix for grid cell transitions.

    Args:
        grid_indices1D (:class:`np.ndarray`)

    Returns:
        M (:class:`numpy.ndarray`)

    """
    if not isinstance(grid_indices1D, np.ndarray):
        raise TypeError(f"Expected np.ndarray, got {type(grid_indices1D)}")

    n = 1 + max(grid_indices1D.flatten())  # number of states

    M = [[0] * n for _ in range(n)]

    for (i, j) in zip(grid_indices1D, grid_indices1D[1:]):
        M[i][j] += 1

    # Convert to probabilities
    for row in M:
        s = sum(row)
        if s > 0:
            row[:] = [f / s for f in row]
    return np.array(M)


def _bins_to_tuple(trj, bins: Union[int, Tuple[int, int]] = 10):
    """Returns tuple of x, y bins

    Args:
        trj: Trajectory
        bins: The bin specification:
            If int, the number of bins for the smallest of the two dimensions such that (min(nx,ny)=bins).
            If [int, int], the number of bins in each dimension (nx, ny = bins).

    Returns:
        bins (Sequence[int,int]): Bins (nx, ny)

    """
    if bins is None:
        bins = 10
    if isinstance(bins, int):
        # make aspect equal
        xlim, ylim = _get_xylim(trj)
        aspect = (ylim[1] - ylim[0]) / (xlim[1] - xlim[0])
        if aspect >= 1:
            bins = (bins, int(bins * aspect))
        else:
            bins = (int(bins / aspect), bins)

    assert len(bins) == 2, f"bins should be length 2 but is {len(bins)}"
    return bins


def calculate_flow_angles(grid_indices: np.ndarray):
    """Calculate average flow between grid indices."""

    bins = (grid_indices[:, 0].max(), grid_indices[:, 1].max())

    M = np.empty((bins[1], bins[0]), dtype=np.ndarray)

    for (i, j) in zip(grid_indices, grid_indices[1:]):
        # Account for fact that grid indices uses 1-base indexing
        ix = i[0] - 1
        iy = i[1] - 1
        jx = j[0] - 1
        jy = j[1] - 1

        if np.array_equal(i, j):
            angle = None
        elif ix == jx and iy > jy:  # move towards y origin (down by default)
            angle = 3 * np.pi / 2
        elif ix == jx and iy < jy:  # move towards y origin (up by default)
            angle = np.pi / 2
        elif ix < jx and iy == jy:  # move right
            angle = 0
        elif ix > jx and iy == jy:  # move left
            angle = np.pi
        elif ix > jx and iy > jy:  # move towards y origin (top left)
            angle = 3 * np.pi / 4
        elif ix > jx and iy < jy:  # move away from y origin (bottom left)
            angle = 5 * np.pi / 4
        elif ix < jx and iy < jy:  # move away from y origin (bottom right)
            angle = 7 * np.pi / 4
        elif ix < jx and iy > jy:  # move towards y origin (top right)
            angle = np.pi / 4
        if angle is not None:
            M[iy, ix] = np.append(M[iy, ix], angle)

    U = np.ones_like(M)  # x component of arrow
    V = np.empty_like(M)  # y component of arrow
    for i, row in enumerate(M):
        for j, angles in enumerate(row):
            x = y = 0
            # average_angle = None
            if angles is not None and len(angles) > 1:
                for angle in angles:
                    if angle is None:
                        continue
                    x += np.cos(angle)
                    y += np.sin(angle)
                # average_angle = np.arctan2(y, x)
                U[i, j] = x
                V[i, j] = y
            else:
                U[i, j] = 0
                V[i, j] = 0

    return U.astype(float), V.astype(float)


def _grid_coords1D(grid_indices: np.ndarray):
    """Convert 2D grid indices to 1D indices."""
    if isinstance(grid_indices, pd.DataFrame):
        grid_indices = grid_indices.values
    grid_indices1D = []
    nr_cols = int(grid_indices[:, 0].max()) + 1
    for coord in grid_indices:
        grid_indices1D.append(
            coord[1] * nr_cols + coord[0]
        )  # nr_rows * col_length + nr_cols

    return np.array(grid_indices1D, dtype=int)


def transitions(trj: TrajaDataFrame, **kwargs):
    """Get first-order Markov model for transitions between grid cells.
    
    Args:
        trj (trajectory)
        kwargs: kwargs to :func:`traja.grid_coordinates`
    
    """
    if "xbin" not in trj.columns or "ybin" not in trj.columns:
        grid_indices = grid_coordinates(trj, **kwargs)
    else:
        grid_indices = trj[["xbin", "ybin"]]

    # Drop nan for converting to int
    grid_indices.dropna(subset=["xbin", "ybin"], inplace=True)
    grid_indices1D = _grid_coords1D(grid_indices)
    transitions_matrix = transition_matrix(grid_indices1D)
    return transitions_matrix


def grid_coordinates(
    trj: TrajaDataFrame,
    bins: Union[int, tuple] = None,
    xlim: tuple = None,
    ylim: tuple = None,
    assign: bool = False,
):
    """Returns ``DataFrame`` of trajectory discretized into 2D lattice grid coordinates.
    Args:
        trj (~`traja.frame.TrajaDataFrame`): Trajectory
        bins (tuple or int)
        xlim (tuple)
        ylim (tuple)
        assign (bool): Return updated original dataframe

    Returns:
        trj (~`traja.frame.TrajaDataFrame`): Trajectory is assign=True otherwise pd.DataFrame

    """
    # Drop nan for converting to int
    trj.dropna(subset=["x", "y"], inplace=True)

    xmin = trj.x.min() if xlim is None else xlim[0]
    xmax = trj.x.max() if xlim is None else xlim[1]
    ymin = trj.y.min() if ylim is None else ylim[0]
    ymax = trj.y.max() if ylim is None else ylim[1]

    bins = _bins_to_tuple(trj, bins)

    if not xlim:
        xbin = pd.cut(trj.x, bins[0], labels=False)
    else:
        xmin, xmax = xlim
        xbinarray = np.linspace(xmin, xmax, bins[0])
        xbin = np.digitize(trj.x, xbinarray)
    if not ylim:
        ybin = pd.cut(trj.y, bins[1], labels=False)
    else:
        ymin, ymax = ylim
        ybinarray = np.linspace(ymin, ymax, bins[1])
        ybin = np.digitize(trj.y, ybinarray)

    if assign:
        trj["xbin"] = xbin
        trj["ybin"] = ybin
        return trj
    return pd.DataFrame({"xbin": xbin, "ybin": ybin})


def generate(
    n: int = 1000,
    random: bool = True,
    step_length: int = 2,
    angular_error_sd: float = 0.5,
    angular_error_dist: Callable = None,
    linear_error_sd: float = 0.2,
    linear_error_dist: Callable = None,
    fps: float = 50,
    spatial_units: str = "m",
    seed: int = None,
    **kwargs,
):
    """Generates a trajectory.

    If ``random`` is ``True``, the trajectory will
    be a correlated random walk/idiothetic directed walk (Kareiva & Shigesada,
    1983), corresponding to an animal navigating without a compass (Cheung,
    Zhang, Stricker, & Srinivasan, 2008). If ``random`` is ``False``, it
    will be a directed walk/allothetic directed walk/oriented path, corresponding
    to an animal navigating with a compass (Cheung, Zhang, Stricker, &
    Srinivasan, 2007, 2008).

    By default, for both random and directed walks, errors are normally
    distributed, unbiased, and independent of each other, so are **simple
    directed walks** in the terminology of Cheung, Zhang, Stricker, & Srinivasan,
    (2008). This behaviour may be modified by specifying alternative values for
    the ``angular_error_dist`` and/or ``linear_error_dist`` parameters.

    The initial angle (for a random walk) or the intended direction (for a
    directed walk) is ``0`` radians. The starting position is ``(0, 0)``.

    Args:
      n (int):  (Default value = 1000)
      random (bool):  (Default value = True)
      step_length:  (Default value = 2)
      angular_error_sd (float):  (Default value = 0.5)
      angular_error_dist (Callable):  (Default value = None)
      linear_error_sd (float):  (Default value = 0.2)
      linear_error_dist (Callable):  (Default value = None)
      fps (float):  (Default value = 50)
      spatial_units:  (Default value = 'm')
      **kwargs: Additional arguments

    Returns:
        trj (:class:`traja.frame.TrajaDataFrame`): Trajectory

    .. note::

        Based on Jim McLean's `trajr <https://github.com/JimMcL/trajr>`_, ported to Python.

        **Reference**: McLean, D. J., & Skowron Volponi, M. A. (2018). trajr: An R package for characterisation of animal
        trajectories. Ethology, 124(6), 440-448. https://doi.org/10.1111/eth.12739.

    """
    if seed is None:
        np.random.seed(0)
    else:
        np.random.seed(seed)
    if angular_error_dist is None:
        angular_error_dist = np.random.normal(
            loc=0.0, scale=angular_error_sd, size=n - 1
        )
    if linear_error_dist is None:
        linear_error_dist = np.random.normal(loc=0.0, scale=linear_error_sd, size=n - 1)
    angular_errors = angular_error_dist
    linear_errors = linear_error_dist
    step_lengths = step_length + linear_errors

    # Don't allow negative lengths
    step_lengths[step_lengths < 0] = 0
    steps = polar_to_z(step_lengths, angular_errors)

    if random:
        # Accumulate angular errors
        coords = np.zeros(n, dtype=np.complex)
        angle = 0
        for i in range(n - 1):
            angle += angular_errors[i]
            length = step_length + linear_errors[i]
            coords[i + 1] = coords[i] + polar_to_z(r=length, theta=angle)
    else:
        coords = np.append(complex(0), np.cumsum(steps))

    x = coords.real
    y = coords.imag

    df = traja.TrajaDataFrame(data={"x": x, "y": y})

    if fps in (0, None):
        raise Exception("fps must be greater than 0")

    df.fps = fps
    time = df.index / fps
    df["time"] = time
    df.spatial_units = spatial_units

    for key, value in kwargs.items():
        df.__dict__[key] = value

    # Update metavars
    metavars = dict(angular_error_sd=angular_error_sd, linear_error_sd=linear_error_sd)
    df.__dict__.update(metavars)

    return df


def _resample_time(
    trj: TrajaDataFrame, step_time: Union[float, int, str], errors="coerce"
):
    if not is_datetime_or_timedelta_dtype(trj.index):
        raise Exception(f"{trj.index.dtype} is not datetime or timedelta.")
    try:
        df = trj.resample(step_time).interpolate(method="spline", order=2)
    except ValueError as e:
        if len(e.args) > 0 and "cannot reindex from a duplicate axis" in e.args[0]:
            if errors == "coerce":
                print("Error: duplicate time indices, keeping first")
                trj = trj.loc[~trj.index.duplicated(keep="first")]
                df = (
                    trj.resample(step_time)
                    .bfill(limit=1)
                    .interpolate(method="spline", order=2)
                )
            else:
                print("Error: duplicate time indices")
                raise ValueError("Duplicate values in indices")
    return df


def resample_time(trj: TrajaDataFrame, step_time: str, new_fps: Optional[bool] = None):
    """Returns a ``TrajaDataFrame`` resampled to consistent `step_time` intervals.

    ``step_time`` should be expressed as a number-time unit combination, eg "2S" for 2 seconds and “2100L” for 2100 milliseconds.

    Args:
        trj (:class:`~traja.frame.TrajaDataFrame`): Trajectory
        step_time (str): step time interval / offset string (eg, '2S' (seconds), '50L' (milliseconds), '50N' (nanoseconds))
        new_fps (bool, optional): new fps

    Results:
        trj (:class:`~traja.frame.TrajaDataFrame`): Trajectory


    .. doctest::

        >>> from traja import generate, resample_time
        >>> df = generate()
        >>> resampled = resample_time(df, '50L') # 50 milliseconds
        >>> resampled.head()
           time          x         y
        0  0.00   0.000000  0.000000
        1  0.05   0.999571  4.293384
        2  0.10  -1.298510  5.423373
        3  0.15  -6.056916  4.874502
        4  0.20 -10.347759  2.108385

    """
    time_col = _get_time_col(trj)
    if time_col == "index" and is_datetime64_any_dtype(trj.index):
        _trj = _resample_time(trj, step_time)
    elif time_col == "index" and is_timedelta64_dtype(trj.index):
        trj.index = pd.to_datetime(trj.index)
        _trj = _resample_time(trj, step_time)
        _trj.index = pd.to_timedelta(_trj.index)
    elif time_col:
        if isinstance(step_time, str):
            try:
                if "." in step_time:
                    raise NotImplementedError(
                        """Fractional step time not implemented.
                          For milliseconds/microseconds/nanoseconds use:
                            L       milliseonds
                            U       microseconds
                            N       nanoseconds
                            eg, step_time='2100L'"""
                    )
            except Exception:
                raise NotImplementedError(
                    f"Inferring from time format {step_time} not yet implemented."
                )
        _trj = trj.set_index(time_col)
        time_units = _trj.__dict__.get("time_units", "s")
        _trj.index = pd.to_datetime(_trj.index, unit=time_units)
        _trj = _resample_time(_trj, step_time)
    else:
        raise NotImplementedError(
            f"Time column ({time_col}) not of expected datasets type."
        )
    return _trj


def rotate(df, angle: Union[float, int] = 0, origin: tuple = None):
    """Returns a ``TrajaDataFrame`` Rotate a trajectory `angle` in radians.

    Args:
        trj (:class:`traja.frame.TrajaDataFrame`): Trajectory
        angle (float): angle in radians
        origin (tuple. optional): rotate around point (x,y)

    Returns:
        trj (:class:`traja.frame.TrajaDataFrame`): Trajectory

    .. note::

        Based on Lyle Scott's `implementation <https://gist.github.com/LyleScott/e36e08bfb23b1f87af68c9051f985302>`_.

    """
    trj = df.copy()
    # Calculate current orientation
    if isinstance(trj, traja.TrajaDataFrame):
        xy = df.traja.xy
    elif isinstance(trj, pd.DataFrame):
        trj = df[["x", "y"]]

    x, y = np.split(xy, [-1], axis=1)
    if origin is None:
        # Assume middle of x and y is origin
        origin = ((x.max() - x.min()) / 2, (y.max() - y.min()) / 2)

    offset_x, offset_y = origin
    new_coords = []

    for x, y in xy:
        adjusted_x = x - offset_x
        adjusted_y = y - offset_y
        cos_rad = math.cos(angle)
        sin_rad = math.sin(angle)
        qx = offset_x + cos_rad * adjusted_x + sin_rad * adjusted_y
        qy = offset_y + -sin_rad * adjusted_x + cos_rad * adjusted_y
        new_coords.append((qx, qy))

    new_xy = np.array(new_coords)
    x, y = np.split(new_xy, [-1], axis=1)
    trj["x"] = x
    trj["y"] = y
    return trj


def rediscretize_points(trj: TrajaDataFrame, R: Union[float, int], time_out=False):
    """Returns a ``TrajaDataFrame`` rediscretized to a constant step length `R`.

    Args:
      trj (:class:`traja.frame.TrajaDataFrame`): Trajectory
      R (float): Rediscretized step length (eg, 0.02)
      time_out (bool): Include time corresponding to time intervals in output

    Returns:
      rt (:class:`numpy.ndarray`): rediscretized trajectory

    """
    if not isinstance(R, (float, int)):
        raise TypeError(f"R should be float or int, but is {type(R)}")

    results = _rediscretize_points(trj, R, time_out)
    rt = results["rt"]
    if len(rt) < 2:
        raise RuntimeError(
            f"Step length {R} is too large for path (path length {len(trj)})"
        )
    rt = traja.from_xy(rt)
    if time_out:
        rt["time"] = results["time"]
    return rt


def _rediscretize_points(
    trj: TrajaDataFrame, R: Union[float, int], time_out=False
) -> dict:
    """Helper function for :func:`traja.trajectory.rediscretize`.

    Args:
      trj (:class:`traja.frame.TrajaDataFrame`): Trajectory
      R (float): Rediscretized step length (eg, 0.02)

    Returns:
      output (dict): Containing:
        result (:class:`numpy.ndarray`): Rediscretized coordinates
        time_vals (optional, list of floats or datetimes): Time points corresponding to result

    """
    # TODO: Implement with complex numbers
    points = trj[["x", "y"]].dropna().values.astype("float64")
    n_points = len(points)
    result = np.empty((128, 2))
    p0 = points[0]
    result[0] = p0
    step_nr = 0
    candidate_start = 1  # running index of candidate

    time_vals = []
    if time_out:
        time_col = _get_time_col(trj)
        time = trj[time_col][0]
        time_vals.append(time)

    while candidate_start <= n_points:
        # Find the first point `curr_ind` for which |points[curr_ind] - p_0| >= R
        curr_ind = np.NaN
        for i in range(
            candidate_start, n_points
        ):  # range of search space for next point
            d = np.linalg.norm(points[i] - result[step_nr])
            if d >= R:
                curr_ind = i  # curr_ind is in [candidate, n_points)
                if time_out:
                    time = trj[time_col][i]
                    time_vals.append(time)
                break
        if np.isnan(curr_ind):
            # End of path
            break

        # The next point may lie on the same segment
        candidate_start = curr_ind

        # The next point lies on the segment p[k-1], p[k]
        curr_result_x = result[step_nr][0]
        prev_x = points[curr_ind - 1, 0]
        curr_result_y = result[step_nr][1]
        prev_y = points[curr_ind - 1, 1]

        # a = 1 if points[k, 0] <= xk_1 else 0
        lambda_ = np.arctan2(
            points[curr_ind, 1] - prev_y, points[curr_ind, 0] - prev_x
        )  # angle
        cos_l = np.cos(lambda_)
        sin_l = np.sin(lambda_)
        U = (curr_result_x - prev_x) * cos_l + (curr_result_y - prev_y) * sin_l
        V = (curr_result_y - prev_y) * cos_l - (curr_result_x - prev_x) * sin_l

        # Compute distance H between (X_{i+1}, Y_{i+1}) and (x_{k-1}, y_{k-1})
        H = U + np.sqrt(abs(R ** 2 - V ** 2))
        XIp1 = H * cos_l + prev_x
        YIp1 = H * sin_l + prev_y

        # Increase array size progressively to make the code run (significantly) faster
        if len(result) <= step_nr + 1:
            result = np.concatenate((result, np.empty_like(result)))

        # Save the point
        result[step_nr + 1] = np.array([XIp1, YIp1])
        step_nr += 1

    # Truncate result
    result = result[: step_nr + 1]
    output = {"rt": result}
    if time_out:
        output["time"] = time_vals
    return output


def _has_cols(trj: TrajaDataFrame, cols: list):
    """Check if `trj` has `cols`."""
    return set(cols).issubset(trj.columns)


def calc_turn_angle(trj: TrajaDataFrame):
    """Return a ``Series`` of floats with turn angles.

    Args:
      trj (:class:`traja.frame.TrajaDataFrame`): Trajectory

    Returns:
        turn_angle (:class:`~pandas.Series`): Turn angle

    .. doctest::

        >>> df = traja.TrajaDataFrame({'x':[0,1,2],'y':[1,2,3]})
        >>> traja.calc_turn_angle(df)
        0    NaN
        1    NaN
        2    0.0
        Name: turn_angle, dtype: float64

    """
    if "heading" not in trj:
        heading = calc_heading(trj)
    else:
        heading = trj.heading
    turn_angle = heading.diff().rename("turn_angle")
    # Correction for 360-degree angle range
    turn_angle.loc[turn_angle >= 180] -= 360
    turn_angle.loc[turn_angle < -180] += 360
    return turn_angle


def calc_angle(trj: TrajaDataFrame):
    """Returns a ``Series`` with angle between steps as a function of displacement w.r.t x axis.

    Args:
       trj (:class:`~traja.frame.TrajaDataFrame`): Trajectory

    Returns:
      angle (:class:`pandas.Series`): Angle series.

    """
    if not _has_cols(trj, ["dx", "displacement"]):
        displacement = calc_displacement(trj)
    else:
        displacement = trj.displacement

    angle = np.rad2deg(np.arccos(np.abs(trj.x.diff()) / displacement))
    return angle


def calc_displacement(trj: TrajaDataFrame, lag=1):
    """Returns a ``Series`` of ``float`` displacement between consecutive indices.

    Args:
        trj (:class:`~traja.frame.TrajaDataFrame`): Trajectory
        lag (int) : time steps between displacement calculation

    Returns:
      displacement (:class:`pandas.Series`): Displacement series.

    .. doctest::

        >>> df = traja.TrajaDataFrame({'x':[0,1,2],'y':[1,2,3]})
        >>> traja.calc_displacement(df)
        0         NaN
        1    1.414214
        2    1.414214
        Name: displacement, dtype: float64

    """
    displacement = np.sqrt(
        np.power(trj.x.shift(lag) - trj.x, 2) + np.power(trj.y.shift(lag) - trj.y, 2)
    )
    displacement.name = "displacement"
    return displacement


def calc_derivatives(trj: TrajaDataFrame):
    """Returns derivatives ``displacement`` and ``displacement_time`` as DataFrame.

    Args:
      trj (:class:`~traja.frame.TrajaDataFrame`): Trajectory

    Returns:
      derivs (:class:`~pandas.DataFrame`): Derivatives.

    .. doctest::

        >>> df = traja.TrajaDataFrame({'x':[0,1,2],'y':[1,2,3],'time':[0., 0.2, 0.4]})
        >>> traja.calc_derivatives(df)
           displacement  displacement_time
        0           NaN                0.0
        1      1.414214                0.2
        2      1.414214                0.4

    """

    time_col = _get_time_col(trj)
    if time_col is None:
        raise Exception("Missing time information in trajectory.")

    if not "displacement" in trj:
        displacement = calc_displacement(trj)
    else:
        displacement = trj.displacement

    # get cumulative seconds
    if is_datetime64_any_dtype(trj[time_col]):
        displacement_time = (
            trj[time_col].astype(int).div(10 ** 9).diff().fillna(0).cumsum()
        )
    else:
        try:
            displacement_time = trj[time_col].diff().fillna(0).cumsum()
        except TypeError:
            raise Exception(
                f"Format (example {trj[time_col][0]}) not recognized as datetime"
            )

    # TODO: Create DataFrame directly
    derivs = pd.DataFrame(
        OrderedDict(displacement=displacement, displacement_time=displacement_time)
    )

    return derivs


def calc_heading(trj: TrajaDataFrame):
    """Calculate trajectory heading.

    Args:
      trj (:class:`~traja.frame.TrajaDataFrame`): Trajectory

    Returns:
        heading (:class:`pandas.Series`): heading as a ``Series``

    ..doctest::

        >>> df = traja.TrajaDataFrame({'x':[0,1,2],'y':[1,2,3]})
        >>> traja.calc_heading(df)
        0     NaN
        1    45.0
        2    45.0
        Name: heading, dtype: float64

    """
    if not _has_cols(trj, ["angle"]):
        angle = calc_angle(trj)
    else:
        angle = trj.angle

    dx = trj.x.diff()
    dy = trj.y.diff()
    # Get heading from angle
    mask = (dx > 0) & (dy >= 0)
    trj.loc[mask, "heading"] = angle[mask]
    mask = (dx >= 0) & (dy < 0)
    trj.loc[mask, "heading"] = -angle[mask]
    mask = (dx < 0) & (dy <= 0)
    trj.loc[mask, "heading"] = -(180 - angle[mask])
    mask = (dx <= 0) & (dy > 0)
    trj.loc[mask, "heading"] = 180 - angle[mask]
    return trj.heading


def speed_intervals(
    trj: TrajaDataFrame, faster_than: float = None, slower_than: float = None
) -> pd.DataFrame:
    """Calculate speed time intervals.

    Returns a dictionary of time intervals where speed is slower and/or faster than specified values.

    Args:
      faster_than (float, optional): Minimum speed threshold. (Default value = None)
      slower_than (float or int, optional): Maximum speed threshold. (Default value = None)

    Returns:
      result (:class:`~pd.DataFrame`) -- time intervals as dataframe

    .. note::

        Implementation ported to Python, heavily inspired by Jim McLean's trajr package.

    .. doctest::

        >> df = traja.generate()
        >> intervals = traja.speed_intervals(df, faster_than=100)
        >> intervals.head()
           start_frame  start_time  stop_frame  stop_time  duration
        0            1        0.02           3       0.06      0.04
        1            4        0.08           8       0.16      0.08
        2           10        0.20          11       0.22      0.02
        3           12        0.24          15       0.30      0.06
        4           17        0.34          18       0.36      0.02

    """
    derivs = get_derivatives(trj)

    if faster_than is None and slower_than is None:
        raise Exception(
            "Parameters faster_than and slower_than are both None, at least one must be provided."
        )

    # Calculate trajectory speeds
    speed = derivs["speed"].values
    times = derivs["speed_times"].values
    times[0] = 0.0
    flags = np.full(len(speed), 1)

    if faster_than is not None:
        flags = flags & (speed > faster_than)
    if slower_than is not None:
        flags = flags & (speed < slower_than)

    changes = np.diff(flags)
    stop_frames = np.where(changes == -1)[0]
    start_frames = np.where(changes == 1)[0]

    # Handle situation where interval begins or ends outside of trajectory
    if len(start_frames) > 0 or len(stop_frames) > 0:
        # Assume interval started at beginning of trajectory, since we don't know what happened before that
        if len(stop_frames) > 0 and (
            len(start_frames) == 0 or stop_frames[0] < start_frames[0]
        ):
            start_frames = np.append(1, start_frames)
        # Similarly, assume that interval can't extend past end of trajectory
        if (
            len(stop_frames) == 0
            or start_frames[len(start_frames) - 1] > stop_frames[len(stop_frames) - 1]
        ):
            stop_frames = np.append(stop_frames, len(speed) - 1)

    stop_times = times[stop_frames]
    start_times = times[start_frames]

    durations = stop_times - start_times
    result = traja.TrajaDataFrame(
        OrderedDict(
            start_frame=start_frames,
            start_time=start_times,
            stop_frame=stop_frames,
            stop_time=stop_times,
            duration=durations,
        )
    )
    return result


def get_derivatives(trj: TrajaDataFrame):
    """Returns derivatives ``displacement``, ``displacement_time``, ``speed``, ``speed_times``, ``acceleration``,
    ``acceleration_times`` as dictionary.

    Args:
        trj (:class:`~traja.frame.TrajaDataFrame`): Trajectory

    Returns:
      derivs (:class:`~pd.DataFrame`) : Derivatives

    .. doctest::

        >> df = traja.TrajaDataFrame({'x':[0,1,2],'y':[1,2,3],'time':[0.,0.2,0.4]})
        >> df.traja.get_derivatives() #doctest: +SKIP
           displacement  displacement_time     speed  speed_times  acceleration  acceleration_times
        0           NaN                0.0       NaN          NaN           NaN                 NaN
        1      1.414214                0.2  7.071068          0.2           NaN                 NaN
        2      1.414214                0.4  7.071068          0.4           0.0                 0.4

    """
    if not _has_cols(trj, ["displacement", "displacement_time"]):
        derivs = calc_derivatives(trj)
        d = derivs["displacement"]
        t = derivs["displacement_time"]
    else:
        d = trj.displacement
        t = trj.displacement_time
        derivs = OrderedDict(displacement=d, displacement_time=t)
    if is_datetime_or_timedelta_dtype(t):
        # Convert to float divisible series
        # TODO: Add support for other time units
        t = t.dt.total_seconds()
    v = d[1 : len(d)] / t.diff()
    v.rename("speed")
    vt = t[1 : len(t)].rename("speed_times")
    # Calculate linear acceleration
    a = v.diff() / vt.diff().rename("acceleration")
    at = vt[1 : len(vt)].rename("accleration_times")

    data = dict(speed=v, speed_times=vt, acceleration=a, acceleration_times=at)
    derivs = derivs.merge(pd.DataFrame(data), left_index=True, right_index=True)

    # Replace infinite values
    derivs.replace([np.inf, -np.inf], np.nan)
    return derivs


def _get_xylim(trj: TrajaDataFrame) -> Tuple[Tuple, Tuple]:
    if (
        "xlim" in trj.__dict__
        and "ylim" in trj.__dict__
        and isinstance(trj.xlim, (list, tuple))
    ):
        return trj.xlim, trj.ylim
    else:
        xlim = trj.x.min(), trj.x.max()
        ylim = trj.y.min(), trj.y.max()
        return xlim, ylim


def coords_to_flow(trj: TrajaDataFrame, bins: Union[int, tuple] = None):
    """Calculate grid cell flow from trajectory.

    Args:
        trj (trajectory)
        bins (int or tuple)

    Returns:
        X (:class:`~numpy.ndarray`): X coordinates of arrow locations
        Y (:class:`~numpy.ndarray`): Y coordinates of arrow locations
        U (:class:`~numpy.ndarray`): X component of vector datasets
        V (:class:`~numpy.ndarray`): Y component of vector datasets

    """
    xlim, ylim = _get_xylim(trj)
    bins = _bins_to_tuple(trj, bins)

    X, Y = np.meshgrid(
        np.linspace(trj.x.min(), trj.x.max(), bins[0]),
        np.linspace(trj.y.min(), trj.y.max(), bins[1]),
    )

    if "xbin" not in trj.columns or "ybin" not in trj.columns:
        grid_indices = traja.grid_coordinates(trj, bins=bins, xlim=xlim, ylim=ylim)
    else:
        grid_indices = trj[["xbin", "ybin"]]

    U, V = traja.calculate_flow_angles(grid_indices.values)

    return X, Y, U, V


def from_xy(xy: np.ndarray):
    """Convenience function for initializing :class:`~traja.frame.TrajaDataFrame` with x,y coordinates.

    Args:
        xy (:class:`numpy.ndarray`): x,y coordinates

    Returns:
        traj_df (:class:`~traja.frame.TrajaDataFrame`): Trajectory as dataframe

    .. doctest::

        >>> import numpy as np
        >>> xy = np.array([[0,1],[1,2],[2,3]])
        >>> traja.from_xy(xy)
           x  y
        0  0  1
        1  1  2
        2  2  3

    """
    df = traja.TrajaDataFrame.from_records(xy, columns=["x", "y"])
    return df


def fill_in_traj(trj: TrajaDataFrame):
    # FIXME: Implement
    return trj


def _get_time_col(trj: TrajaDataFrame):
    # Check if saved in metadata
    time_col = trj.__dict__.get("time_col", None)
    if time_col:
        return time_col
    # Check if index is datetime
    if is_datetime64_any_dtype(trj.index) or is_datetime_or_timedelta_dtype(trj.index):
        return "index"
    # Check if any column contains 'time'
    time_cols = [col for col in trj if "time" in col.lower()]
    if time_cols:
        # Try first column
        time_col = time_cols[0]
        if is_datetime_or_timedelta_dtype(trj[time_col]):
            return time_col
        else:
            # Time column is float, etc. but not datetime64.
            # FIXME: Add conditional return, etc.
            return time_col
    else:
        # No time column found
        return None


if __name__ == "__main__":
    import doctest

    doctest.testmod()



import copy
import logging
from typing import Optional, List, Union, Tuple

import numpy as np
import pandas as pd
from pandas import DataFrame
from pandas.api.types import is_numeric_dtype

import traja

logging.basicConfig(format="%(levelname)s:%(message)s", level=logging.ERROR)


class TrajaDataFrame(pd.DataFrame):
    """A TrajaDataFrame object is a subclass of pandas :class:`<~pandas.dataframe.DataFrame>`.

    Args:
      args: Typical arguments for pandas.DataFrame.

    Returns:
      traja.TrajaDataFrame -- TrajaDataFrame constructor.
      
        >>> traja.TrajaDataFrame({'x':[0,1,2],'y':[1,2,3]}) # doctest: +SKIP
           x  y
        0  0  1
        1  1  2
        2  2  3

    """

    _metadata = [
        "xlim",
        "ylim",
        "spatial_units",
        "xlabel",
        "ylabel",
        "title",
        "fps",
        "time_units",
        "time_col",
        "id",
    ]

    def __init__(self, *args, **kwargs):
        # Allow setting metadata from constructor
        traja_kwargs = dict()
        for key in list(kwargs.keys()):
            for name in self._metadata:
                if key == name:
                    traja_kwargs[key] = kwargs.pop(key)
        super(TrajaDataFrame, self).__init__(*args, **kwargs)

        if len(args) == 1 and isinstance(args[0], TrajaDataFrame):
            args[0]._copy_attrs(self)
        for name, value in traja_kwargs.items():
            self.__dict__[name] = value

        # Initialize metadata like 'fps','spatial_units', etc.
        self._init_metadata()

    @property
    def _constructor(self):
        return TrajaDataFrame

    def _copy_attrs(self, df):
        for attr in self._metadata:
            df.__dict__[attr] = getattr(self, attr, None)

    def __finalize__(self, other, method=None, **kwargs):
        """propagate metadata from other to self """
        # merge operation: using metadata of the left object
        if method == "merge":
            for name in self._metadata:
                object.__setattr__(self, name, getattr(other.left, name, None))
        # concat operation: using metadata of the first object
        elif method == "concat":
            for name in self._metadata:
                object.__setattr__(self, name, getattr(other.objs[0], name, None))
        else:
            for name in self._metadata:
                object.__setattr__(self, name, getattr(other, name, None))
        return self

    # def __getitem__(self, key):
    #     """
    #       If result is a DataFrame with a x or X column, return a
    #       TrajaDataFrame.
    #       """
    #     result = super(TrajaDataFrame, self).__getitem__(key)
    #     if isinstance(result, DataFrame) and "x" == result or "X" == result:
    #         result.__class__ = TrajaDataFrame
    #     elif isinstance(result, DataFrame):
    #         result.__class__ = DataFrame
    #     return result

    def _init_metadata(self):
        defaults = dict(fps=None, spatial_units="m", time_units="s")
        for name, value in defaults.items():
            if name not in self.__dict__:
                self.__dict__[name] = value

    def _get_time_col(self):
        time_cols = [col for col in self if "time" in col.lower()]
        if time_cols:
            time_col = time_cols[0]
            if is_numeric_dtype(self[time_col]):
                return time_col
        else:
            return None

    @classmethod
    def from_xy(cls, xy: np.ndarray):
        """Convenience function for initializing :class:`~traja.frame.TrajaDataFrame` with x,y coordinates.

        Args:
            xy (:class:`numpy.ndarray`): x,y coordinates

        Returns:
            traj_df (:class:`~traja.frame.TrajaDataFrame`): Trajectory as dataframe

        .. doctest::

            >>> import numpy as np
            >>> xy = np.array([[0,1],[1,2],[2,3]])
            >>> traja.from_xy(xy)
               x  y
            0  0  1
            1  1  2
            2  2  3

        """
        df = cls.from_records(xy, columns=["x", "y"])
        return df

    def set(self, key, value):
        """Set metadata."""
        self.__dict__[key] = value


def tocontainer(func):
    def wrapper(*args, **kwargs):
        result = func(*args, **kwargs)
        return TrajaCollection(result)

    return wrapper


class TrajaCollection(TrajaDataFrame):
    """Collection of trajectories."""

    _metadata = [
        "xlim",
        "ylim",
        "spatial_units",
        "xlabel",
        "ylabel",
        "title",
        "fps",
        "time_units",
        "time_col",
        "_id_col",
    ]

    def __init__(
        self,
        trjs: Union[TrajaDataFrame, pd.DataFrame, dict],
        id_col: Optional[str] = None,
        **kwargs,
    ):
        """Initialize with trajectories with x, y, and time columns.

        Args:self.
            trjs
            id_col (str) - Default is "id"

        """
        # Add id column
        if isinstance(trjs, dict):
            _trjs = []
            for name, df in trjs.items():
                df["id"] = name
                _trjs.append(df)
            super(TrajaCollection, self).__init__(pd.concat(_trjs), **kwargs)
        elif isinstance(trjs, (TrajaDataFrame, DataFrame)):
            super(TrajaCollection, self).__init__(trjs, **kwargs)
        else:
            super(TrajaCollection, self).__init__(trjs, **kwargs)

        if id_col:
            self._id_col = id_col
        elif hasattr(self, "_id_col"):
            self._id_col = self._id_col
        else:
            self._id_col = "id"  # default

    @property
    def _constructor(self):
        return TrajaCollection

    def _copy_attrs(self, df):
        for attr in self._metadata:
            df.__dict__[attr] = getattr(self, attr, None)

    # def __copy__(self):
    #     return TrajaCollection(self.trjs).__dict__.update(self.__dict__)

    def __repr__(self):
        return "TrajaCollection:\n" + super(TrajaCollection, self).__repr__()

    # def __add__(self, other):
    #     trjs = self.trjs.append(other, ignore_index=True)
    #     return TrajaCollection(trjs, id_col=self._id_col)

    def plot(self, colors=None, **kwargs):
        """Plot collection of trajectories with colors assigned to each id.

            >>> trjs = {ind: traja.generate(seed=ind) for ind in range(3)} # doctest: +SKIP
            >>> coll = traja.TrajaCollection(trjs) # doctest: +SKIP
            >>> coll.plot() # doctest: +SKIP

        """
        return traja.plotting.plot_collection(
            self, self._id_col, colors=colors, **kwargs
        )

    def apply_all(self, method, **kwargs):
        """Applies method to all trajectories

        Args:
            method

        Returns:
            dataframe or series

            >>> trjs = {ind: traja.generate(seed=ind) for ind in range(3)} # doctest: +SKIP
            >>> coll = traja.TrajaCollection(trjs) # doctest: +SKIP
            >>> angles = coll.apply_all(traja.calc_angles) # doctest: +SKIP

        """
        return self.groupby(by=self._id_col).apply(method)


class StaticObject(object):
    def __init__(
        self,
        x: Optional[float] = None,
        y: Optional[float] = None,
        bounding_box: Tuple[float] = None,
    ):
        ...
        pass



from typing import Union

import pandas as pd
from pandas.api.types import is_datetime64_any_dtype

import traja


@pd.api.extensions.register_dataframe_accessor("traja")
class TrajaAccessor(object):
    """Accessor for pandas DataFrame with trajectory-specific numerical and analytical functions.

    Access with `df.traja`."""

    def __init__(self, pandas_obj):
        self._validate(pandas_obj)
        self._obj = pandas_obj

    def _strip(self, text):
        try:
            return text.strip()
        except AttributeError:
            return pd.to_numeric(text, errors="coerce")

    @staticmethod
    def _validate(obj):
        if "x" not in obj.columns or "y" not in obj.columns:
            raise AttributeError("Must have 'x' and 'y'.")

    @property
    def center(self):
        """Return the center point of this trajectory."""
        x = self._obj.x
        y = self._obj.y
        return (float(x.mean()), float(y.mean()))

    @property
    def bounds(self):
        """Return limits of x and y dimensions (``(xmin, xmax), (ymin, ymax)``)."""
        xlim = self._obj.x.min(), self._obj.x.max()
        ylim = self._obj.y.min(), self._obj.y.max()
        return (xlim, ylim)

    def night(self, begin: str = "19:00", end: str = "7:00"):
        """Get nighttime datasets between `begin` and `end`.

        Args:
          begin (str):  (Default value = '19:00')
          end (str):  (Default value = '7:00')

        Returns:
          trj (:class:`~traja.frame.TrajaDataFrame`): Trajectory during night.

        """
        return self.between(begin, end)

    def day(self, begin: str = "7:00", end: str = "19:00"):
        """Get daytime datasets between `begin` and `end`.

        Args:
          begin (str):  (Default value = '7:00')
          end (str):  (Default value = '19:00')

        Returns:
          trj (:class:`~traja.frame.TrajaDataFrame`): Trajectory during day.

        """
        return self.between(begin, end)

    def _get_time_col(self):
        """Returns time column in trajectory.

        Args:

        Returns:
           time_col (str or None): name of time column, 'index' or None

        """
        return traja.trajectory._get_time_col(self._obj)

    def between(self, begin: str, end: str):
        """Returns trajectory between `begin` and end` if `time` column is `datetime64`.

        Args:
          begin (str): Beginning of time slice.
          end (str): End of time slice.

        Returns:
          trj (:class:`~traja.frame.TrajaDataFrame`): Dataframe between values.
          
        .. doctest ::

            >>> s = pd.to_datetime(pd.Series(['Jun 30 2000 12:00:01', 'Jun 30 2000 12:00:02', 'Jun 30 2000 12:00:03']))
            >>> df = traja.TrajaDataFrame({'x':[0,1,2],'y':[1,2,3],'time':s})
            >>> df.traja.between('12:00:00','12:00:01')
                             time  x  y
            0 2000-06-30 12:00:01  0  1

        """
        time_col = self._get_time_col()
        if time_col == "index":
            return self._obj.between_time(begin, end)
        elif time_col and is_datetime64_any_dtype(self._obj[time_col]):
            # Backup index
            dt_index_col = self._obj.index.name
            # Set dt_index
            trj = self._obj.copy()
            trj.set_index(time_col, inplace=True)
            # Create slice of trajectory
            trj = trj.between_time(begin, end)
            # Restore index and return column
            if dt_index_col:
                trj.set_index(dt_index_col, inplace=True)
            else:
                trj.reset_index(inplace=True)
            return trj
        else:
            raise TypeError("Either time column or index must be datetime64")

    def resample_time(self, step_time: float):
        """Returns trajectory resampled with ``step_time``.

        Args:
           step_time (float): Step time

        Returns:
            trj (:class:`~traja.frame.TrajaDataFrame`): Dataframe resampled.
        """
        return traja.trajectory.resample_time(self._obj, step_time=step_time)

    def trip_grid(
        self,
        bins: Union[int, tuple] = 10,
        log: bool = False,
        spatial_units=None,
        normalize: bool = False,
        hist_only: bool = False,
        plot: bool = True,
        **kwargs,
    ):
        """Returns a 2D histogram of trip.

        Args:
          bins (int, optional): Number of bins (Default value = 16)
          log (bool): log scale histogram (Default value = False)
          spatial_units (str): units for plotting
          normalize (bool): normalize histogram into density plot
          hist_only (bool): return histogram without plotting

        Returns:
            hist (:class:`numpy.ndarray`): 2D histogram as array
            image (:class:`matplotlib.collections.PathCollection`: image of histogram

        """
        hist, image = traja.plotting.trip_grid(
            self._obj,
            bins=bins,
            log=log,
            spatial_units=self._obj.get("spatial_units", "m"),
            normalize=normalize,
            hist_only=hist_only,
            plot=plot,
            **kwargs,
        )
        return hist, image

    def plot(self, n_coords: int = None, show_time=False, **kwargs):
        """Plot trajectory over period.

        Args:
          n_coords (int): Number of coordinates to plot
          **kwargs: additional keyword arguments to :meth:`matplotlib.axes.Axes.scatter`

        Returns:
            ax (:class:`~matplotlib.axes.Axes`): Axes of plot
        """
        ax = traja.plotting.plot(
            trj=self._obj,
            accessor=self,
            n_coords=n_coords,
            show_time=show_time,
            **kwargs,
        )
        return ax

    def plot_3d(self, **kwargs):
        """Plot 3D trajectory for single identity over period.

        Args:
        trj (:class:`traja.TrajaDataFrame`): trajectory
        n_coords (int, optional): Number of coordinates to plot
        **kwargs: additional keyword arguments to :meth:`matplotlib.axes.Axes.scatter`

        Returns:
            collection (:class:`~matplotlib.collections.PathCollection`): collection that was plotted

        .. note::
            Takes a while to plot large trajectories. Consider using first::
            
                rt = trj.traja.rediscretize(R=1.) # Replace R with appropriate step length
                rt.traja.plot_3d()

        """
        ax = traja.plotting.plot_3d(trj=self._obj, **kwargs)
        return ax

    def plot_flow(self, kind="quiver", **kwargs):
        """Plot grid cell flow.

        Args:
          kind (str): Kind of plot (eg, 'quiver','surface','contour','contourf','stream')
          **kwargs: additional keyword arguments to :meth:`matplotlib.axes.Axes.scatter`

        Returns:
            ax (:class:`~matplotlib.axes.Axes`): Axes of plot

        """
        ax = traja.plotting.plot_flow(trj=self._obj, kind=kind, **kwargs)
        return ax

    def plot_collection(self, colors=None, **kwargs):
        return traja.plotting.plot_collection(
            self._obj, id_col=self._id_col, colors=colors, **kwargs
        )

    def apply_all(self, method, id_col=None, **kwargs):
        """Applies method to all trajectories and returns grouped dataframes or series"""
        id_col = id_col or getattr(self, "_id_col", "id")
        return self._obj.groupby(by=id_col).apply(method, **kwargs)

    def _has_cols(self, cols: list):
        return traja.trajectory._has_cols(self._obj, cols)

    @property
    def xy(self):
        """Returns a :class:`numpy.ndarray` of x,y coordinates.

        Args:
            split (bool): Split into seaprate x and y :class:`numpy.ndarrays`

        Returns:
          xy (:class:`numpy.ndarray`) -- x,y coordinates (separate if `split` is `True`)
          
        .. doctest::

            >>> df = traja.TrajaDataFrame({'x':[0,1,2],'y':[1,2,3]})
            >>> df.traja.xy
            array([[0, 1],
                   [1, 2],
                   [2, 3]])

        """
        if self._has_cols(["x", "y"]):
            xy = self._obj[["x", "y"]].values
            return xy
        else:
            raise Exception("'x' and 'y' are not in the dataframe.")

    def _check_has_time(self):
        """Check for presence of displacement time column."""
        time_col = self._get_time_col()
        if time_col is None:
            raise Exception("Missing time information in trajectory.")

    def transitions(self, *args, **kwargs):
        """Calculate transition matrix"""
        return traja.transitions(self._obj, *args, **kwargs)

    def calc_derivatives(self, assign: bool = False):
        """Returns derivatives `displacement` and `displacement_time`.

        Args:
          assign (bool): Assign output to ``TrajaDataFrame`` (Default value = False)

        Returns:
          derivs (:class:`~collections.OrderedDict`): Derivatives.
          
        .. doctest::

            >>> df = traja.TrajaDataFrame({'x':[0,1,2],'y':[1,2,3],'time':[0., 0.2, 0.4]})
            >>> df.traja.calc_derivatives()
               displacement  displacement_time
            0           NaN                0.0
            1      1.414214                0.2
            2      1.414214                0.4


        """
        derivs = traja.trajectory.calc_derivatives(self._obj)
        if assign:
            trj = self._obj.merge(derivs, left_index=True, right_index=True)
            self._obj = trj
        return derivs

    def get_derivatives(self) -> pd.DataFrame:
        """Returns derivatives as DataFrame."""
        derivs = traja.trajectory.get_derivatives(self._obj)
        return derivs

    def speed_intervals(
        self,
        faster_than: Union[float, int] = None,
        slower_than: Union[float, int] = None,
    ):
        """Returns ``TrajaDataFrame`` with speed time intervals.

        Returns a dataframe of time intervals where speed is slower and/or faster than specified values.

        Args:
          faster_than (float, optional): Minimum speed threshold. (Default value = None)
          slower_than (float or int, optional): Maximum speed threshold. (Default value = None)

        Returns:
          result (:class:`~pandas.DataFrame`) -- time intervals as dataframe

        .. note::

            Implementation ported to Python, heavily inspired by Jim McLean's trajr package.

        """
        result = traja.trajectory.speed_intervals(self._obj, faster_than, slower_than)
        return result

    def to_shapely(self):
        """Returns shapely object for area, bounds, etc. functions.

        Args:

        Returns:
          shape (shapely.geometry.linestring.LineString): Shapely shape.
          
        .. doctest::

            >>> df = traja.TrajaDataFrame({'x':[0,1,2],'y':[1,2,3]})
            >>> shape = df.traja.to_shapely()
            >>> shape.is_closed
            False

        """
        trj = self._obj[["x", "y"]].dropna()
        tracks_shape = traja.trajectory.to_shapely(trj)
        return tracks_shape

    def calc_displacement(self, assign: bool = True) -> pd.Series:
        """Returns ``Series`` of `float` with displacement between consecutive indices.

        Args:
          assign (bool, optional): Assign displacement to TrajaAccessor (Default value = True)

        Returns:
          displacement (:class:`pandas.Series`): Displacement series.
          
        .. doctest::

            >>> df = traja.TrajaDataFrame({'x':[0,1,2],'y':[1,2,3]})
            >>> df.traja.calc_displacement()
            0         NaN
            1    1.414214
            2    1.414214
            Name: displacement, dtype: float64

        """
        displacement = traja.trajectory.calc_displacement(self._obj)
        if assign:
            self._obj = self._obj.assign(displacement=displacement)
        return displacement

    def calc_angle(self, assign: bool = True) -> pd.Series:
        """Returns ``Series`` with angle between steps as a function of displacement w.r.t x axis.

        Args:
          assign (bool, optional): Assign turn angle to TrajaAccessor (Default value = True)

        Returns:
          angle (:class:`pandas.Series`): Angle series.
          
        .. doctest::

            >>> df = traja.TrajaDataFrame({'x':[0,1,2],'y':[1,2,3]})
            >>> df.traja.calc_angle()
            0     NaN
            1    45.0
            2    45.0
            dtype: float64

        """
        angle = traja.trajectory.calc_angle(self._obj)
        if assign:
            self._obj["angle"] = angle
        return angle

    def scale(self, scale: float, spatial_units: str = "m"):
        """Scale trajectory when converting, eg, from pixels to meters.

        Args:
          scale(float): Scale to convert coordinates
          spatial_units(str., optional): Spatial units (eg, 'm') (Default value = "m")

        .. doctest::

            >>> df = traja.TrajaDataFrame({'x':[0,1,2],'y':[1,2,3]})
            >>> df.traja.scale(0.1)
            >>> df
                 x    y
            0  0.0  0.1
            1  0.1  0.2
            2  0.2  0.3

        """
        self._obj[["x", "y"]] *= scale
        self._obj.__dict__["spatial_units"] = spatial_units

    def _transfer_metavars(self, df):
        for attr in self._obj._metadata:
            df.__dict__[attr] = getattr(self._obj, attr, None)
        return df

    def rediscretize(self, R: float):
        """Resample a trajectory to a constant step length. R is rediscretized step length.

        Args:
          R (float): Rediscretized step length (eg, 0.02)

        Returns:
          rt (:class:`traja.TrajaDataFrame`): rediscretized trajectory
          
        .. note::
          
            Based on the appendix in Bovet and Benhamou, (1988) and Jim McLean's
            `trajr <https://github.com/JimMcL/trajr>`_ implementation.
          
        .. doctest::

            >>> df = traja.TrajaDataFrame({'x':[0,1,2],'y':[1,2,3]})
            >>> df.traja.rediscretize(1.)
                      x         y
            0  0.000000  1.000000
            1  0.707107  1.707107
            2  1.414214  2.414214

        """
        if not isinstance(R, (int, float)):
            raise ValueError(f"R must be provided as float or int")
        rt = traja.trajectory.rediscretize_points(self._obj, R)
        self._transfer_metavars(rt)
        return rt

    def grid_coordinates(self, **kwargs):
        return traja.grid_coordinates(self._obj, **kwargs)

    def calc_heading(self, assign: bool = True):
        """Calculate trajectory heading.

        Args:
          assign (bool): (Default value = True)

        Returns:
            heading (:class:`pandas.Series`): heading as a ``Series``

        ..doctest::

            >>> df = traja.TrajaDataFrame({'x':[0,1,2],'y':[1,2,3]})
            >>> df.traja.calc_heading()
            0     NaN
            1    45.0
            2    45.0
            Name: heading, dtype: float64

        """
        heading = traja.trajectory.calc_heading(self._obj)
        if assign:
            self._obj["heading"] = heading
        return heading

    def calc_turn_angle(self, assign: bool = True):
        """Calculate turn angle.

        Args:
          assign (bool):  (Default value = True)

        Returns:
            turn_angle (:class:`~pandas.Series`): Turn angle

        .. doctest::

            >>> df = traja.TrajaDataFrame({'x':[0,1,2],'y':[1,2,3]})
            >>> df.traja.calc_turn_angle()
            0    NaN
            1    NaN
            2    0.0
            Name: turn_angle, dtype: float64

        """
        turn_angle = traja.trajectory.calc_turn_angle(self._obj)

        if assign:
            self._obj["turn_angle"] = turn_angle
        return turn_angle



from typing import Optional, Union

import numpy as np
import pandas as pd
from pandas.core.dtypes.common import is_datetime64_any_dtype, is_timedelta64_dtype

from traja import TrajaDataFrame


def from_df(df: pd.DataFrame, xcol=None, ycol=None, time_col=None, **kwargs):
    """Returns a :class:`traja.frame.TrajaDataFrame` from a :class:`pandas DataFrame<pandas.DataFrame>`.

    Args:
      df (:class:`pandas.DataFrame`): Trajectory as pandas ``DataFrame``
      xcol (str)
      ycol (str)
      timecol (str)

    Returns:
      traj_df (:class:`~traja.frame.TrajaDataFrame`): Trajectory

    .. doctest::

        >>> df = pd.DataFrame({'x':[0,1,2],'y':[1,2,3]})
        >>> traja.from_df(df)
           x  y
        0  0  1
        1  1  2
        2  2  3

    """
    traj_df = TrajaDataFrame(df)

    # Identify x and y columns if defined by user
    if xcol and ycol:
        traj_df["x"] = pd.to_numeric(traj_df[xcol], errors="coerce")
        traj_df["y"] = pd.to_numeric(traj_df[ycol], errors="coerce")
    if time_col:
        traj_df[time_col] = pd.to_timedelta(
            traj_df[time_col], unit=kwargs.get("time_units", "s")
        )
        kwargs.update({"time_col": time_col})

    # Initialize metadata
    for var in traj_df._metadata:
        if not hasattr(traj_df, var):
            traj_df.__dict__[var] = None

    # Save additional metadata
    for key, val in kwargs.items():
        traj_df.__dict__[key] = val
    return traj_df


def read_file(
    filepath: str,
    id: Optional[str] = None,
    xcol: Optional[str] = None,
    ycol: Optional[str] = None,
    parse_dates: Union[str, bool] = False,
    xlim: Optional[tuple] = None,
    ylim: Optional[tuple] = None,
    spatial_units: str = "m",
    fps: Optional[float] = None,
    **kwargs,
):
    """Convenience method wrapping pandas `read_csv` and initializing metadata.

    Args:
      filepath (str): path to csv file with `x`, `y` and `time` (optional) columns
      id (str): id for trajectory
      xcol (str): name of column containing x coordinates
      ycol (str): name of column containing y coordinates
      parse_dates (Union[list,bool]): The behavior is as follows:
                                    - boolean. if True -> try parsing the index.
                                    - list of int or names. e.g. If [1, 2, 3] -> try parsing columns 1, 2, 3 each as a
                                    separate date column.
      xlim (tuple): x limits (min,max) for plotting
      ylim (tuple): y limits (min,max) for plotting
      spatial_units (str): for plotting (eg, 'cm')
      fps (float): for time calculations
      **kwargs: Additional arguments for :meth:`pandas.read_csv`.

    Returns:
        traj_df (:class:`~traja.main.TrajaDataFrame`): Trajectory

    """
    date_parser = kwargs.pop("date_parser", None)

    # TODO: Set index to first column containing 'time'
    df_test = pd.read_csv(
        filepath, nrows=10, parse_dates=parse_dates, infer_datetime_format=True
    )

    if xcol is not None or ycol is not None:
        if not xcol in df_test or ycol not in df_test:
            raise Exception(f"{xcol} or {ycol} not found as headers.")

    # Strip whitespace
    whitespace_cols = [c for c in df_test if " " in df_test[c].name]
    stripped_cols = {c: lambda x: x.strip() for c in whitespace_cols}
    converters = {**stripped_cols, **kwargs.pop("converters", {})}

    # Downcast to float32 # TODO: Benchmark float32 vs float64 for very big datasets
    float_cols = df_test.select_dtypes(include=[np.float]).columns
    float32_cols = {c: np.float32 for c in float_cols}

    # Convert string columns to categories
    string_cols = [c for c in df_test if df_test[c].dtype == str]
    category_cols = {c: "category" for c in string_cols}
    dtype = {**float32_cols, **category_cols, **kwargs.pop("dtype", {})}

    # Parse time column if present
    time_cols = [col for col in df_test.columns if "time" in col.lower()]
    time_col = time_cols[0] if time_cols else None

    if parse_dates and not date_parser and time_col:
        # try different parsers
        format_strs = [
            "%Y-%m-%d %H:%M:%S:%f",
            "%Y-%m-%d %H:%M:%S.%f",
            "%Y-%m-%d %H:%M:%S",
        ]
        for format_str in format_strs:
            date_parser = lambda x: pd.datetime.strptime(x, format_str)
            try:
                df_test = pd.read_csv(
                    filepath, date_parser=date_parser, nrows=10, parse_dates=[time_col]
                )
            except ValueError:
                pass
            if is_datetime64_any_dtype(df_test[time_col]):
                break
            elif is_timedelta64_dtype(df_test[time_col]):
                break
            else:
                # No datetime or timestamp column found
                date_parser = None

    if "csv" in filepath:
        trj = pd.read_csv(
            filepath,
            date_parser=date_parser,
            parse_dates=parse_dates or [time_col] if date_parser else False,
            converters=converters,
            dtype=dtype,
            **kwargs,
        )

        # TODO: Replace default column renaming with user option if needed
        if time_col:
            trj.rename(columns={time_col: "time"})
        elif fps is not None:
            time = np.array([x for x in trj.index], dtype=int) / fps
            trj["time"] = time
        else:
            # leave index as int frames
            pass
        if xcol and ycol:
            trj.rename(columns={xcol: "x", ycol: "y"})
    else:
        # TODO: Implement for HDF5 and .npy files.
        raise NotImplementedError("Non-csv's not yet implemented")

    trj = TrajaDataFrame(trj)

    # Set meta properties of TrajaDataFrame
    metadata = dict(
        id=id,
        xlim=xlim,
        spatial_units=spatial_units,
        title=kwargs.get("title", None),
        xlabel=kwargs.get("xlabel", None),
        ylabel=kwargs.get("ylabel", None),
        fps=fps,
    )
    trj.__dict__.update(**metadata)
    return trj



import glob
import pickle

import numpy as np
import matplotlib.pyplot as plt

from sklearn.decomposition import PCA


import traja

files = sorted(
    glob.glob("/Users/justinshenk/neurodata/datasets/raw_centroids_rev2/*.csv")
)
cages = ["E1", "E2", "E4", "F4", "F6"]

xlim = -0.058, 0.058
ylim = -0.119, 0.119

path_grids = {}

for cage in cages:
    path_grids[cage] = {}
    path_grids[cage]["day"] = []
    path_grids[cage]["night"] = []
    cagefiles = [file for file in files if cage in file]
    for idx, file in enumerate(cagefiles):
        if idx % 10 != 0:
            continue
        df = traja.read_file(file, xlim=xlim, ylim=ylim)
        day, _ = df.traja.day().traja.trip_grid(hist_only=True, log=True)
        night, _ = df.traja.night().traja.trip_grid(hist_only=True, log=True)
        path_grids[cage]["day"].append((idx, day))
        path_grids[cage]["night"].append((idx, night))
        print(f"Completed {cage}, index {idx}")
    print(f"Completed {cage}")

# path_grids_dest = "path_grids.pkl"
# pickle.dump(path_grids, open(path_grids_dest, "wb"))

X = []
target_names = []
path_grids = pickle.load(open("traja/path_grids.pkl", "rb"))
cages = path_grids.keys()
for cage in cages:
    for period in ("day", "night"):
        grids = path_grids[cage][period]
        for idx, grid in grids:
            X.append(grid.flatten())
            target_names.append(f"{cage}-{period}-{idx}")

days_from_surgery = set([int(x.split("-")[-1]) for x in target_names])

X = np.array(X)
pca = PCA(n_components=2)
X_r = pca.fit(X).transform(X)

# Group by period
plt.figure()
lw = 2
colors = ["navy", "darkorange"]

# for color, i, target_name in zip(colors, [0, 1], target_names):
day_idx = np.array(["day" in x for x in target_names])

plt.scatter(
    X_r[day_idx, 0], X_r[day_idx, 1], color=f"C0", alpha=0.8, lw=lw, label="Day"
)
plt.scatter(
    X_r[~day_idx, 0], X_r[~day_idx, 1], color=f"C1", alpha=0.8, lw=lw, label="Night"
)
plt.legend(loc="best", shadow=False, scatterpoints=1)
plt.title("PCA of Mouse trajectory dataset")
plt.show()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from setuptools import setup, find_packages

import os
import re

here = os.path.abspath(os.path.dirname(__file__))


def read(*parts):
    with open(os.path.join(here, *parts), "r", encoding="utf8") as fp:
        return fp.read()


# Get package version
def find_version(*file_paths):
    version_file = read(*file_paths)
    version_match = re.search(r"^__version__ = ['\"]([^'\"]*)['\"]", version_file, re.M)
    if version_match:
        return version_match.group(1)
    raise RuntimeError("Unable to find version string.")


requirements = ["matplotlib", "pandas", "numpy", "shapely", "scipy", "tzlocal"]

extras_requirements = {"all": ["torch", "rpy2", "tzlocal", "fastdtw"]}

this_dir = os.path.abspath(os.path.dirname(__file__))
with open(os.path.join(this_dir, "README.rst"), encoding="utf-8") as f:
    long_description = f.read()

setup(
    name="traja",
    version=find_version("traja", "__init__.py"),
    description="Traja is a trajectory analysis and visualization tool",
    url="https://github.com/justinshenk/traja",
    author="Justin Shenk",
    author_email="shenkjustin@gmail.com",
    long_description=long_description,
    long_description_content_type="text/markdown",
    install_requires=requirements,
    extras_require=extras_requirements,
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Intended Audience :: Education",
        "Intended Audience :: Science/Research",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Topic :: Scientific/Engineering :: Mathematics",
        "Topic :: Software Development :: Libraries :: Python Modules",
        "Topic :: Software Development :: Libraries",
    ],
    python_requires=">= 3.6",
    packages=find_packages(),
    include_package_data=True,
    license="MIT",
    keywords="trajectory analysis",
    zip_safe=False,
)



import warnings

import numpy as np
import numpy.testing as npt
from rpy2 import rinterface
from rpy2.rinterface import RRuntimeWarning

import traja
from traja import rutils


warnings.filterwarnings("ignore", category=RRuntimeWarning, module="rpy2")
warnings.filterwarnings("ignore", category=UserWarning, module="rpy2")


df = traja.generate(n=20)


def test_import_adehabitat():
    traja.rutils.import_adehabitat(suppress_messages=True)


def test_import_trajr():
    traja.rutils.import_trajr(suppress_messages=True)


def test_to_trajr():
    trjr = traja.rutils.to_trajr(df)
    assert "x" in trjr
    assert "y" in trjr
    assert "time" in trjr
    assert "polar" in trjr
    assert "displacement" in trjr
    actual = trjr.x[:3].values
    expected = np.array([0.0, 1.162_605_74, 1.861_836_8])

    npt.assert_allclose(actual, expected)


def test_to_ltraj():
    ltraj = traja.rutils.to_ltraj(df)
    rdataframe = ltraj
    assert "x" in rdataframe
    assert "y" in rdataframe
    assert len(rdataframe) == 20



import warnings

import numpy as np
import numpy.testing as npt
import matplotlib

matplotlib.use("Agg")
import matplotlib.pyplot as plt
import pandas as pd

import traja

warnings.filterwarnings("ignore", category=UserWarning, module="matplotlib")

df = traja.generate(n=10)


def test_stylize_axes():
    collection = traja.plot(df, interactive=False)
    traja.plotting.stylize_axes(collection.axes)


def test_color_dark():
    df = traja.generate(n=10)
    index = pd.DatetimeIndex(range(10))
    df.index = index
    traja.plot(df, interactive=False)
    ax = plt.gca()
    traja.color_dark(df.x, ax)


def test_sans_serif():
    traja.plotting.sans_serif()


def test_plot_3d():
    traja.plot_3d(df, interactive=False)


def test_plot_flow():
    traja.plot_flow(df, interactive=False)


def test_plot_contour():
    ax = traja.plot_contour(df, interactive=False)


def test_plot_surface():
    ax = traja.plot_surface(df, interactive=False)


def test_plot_stream():
    ax = traja.plot_stream(df, interactive=False)


def test_trip_grid():
    traja.plotting.trip_grid(df, interactive=False)


def test_label_axes():
    df.traja.plot(interactive=False)
    ax = plt.gca()
    traja.plotting._label_axes(df, ax)


def test_plot_actogram():
    df = traja.generate(n=10)
    index = pd.DatetimeIndex(range(10))
    df.index = index
    activity = traja.calc_displacement(df)
    activity.name = "activity"
    traja.plotting.plot_actogram(df.x, interactive=False)


def test_plot_xy():
    traja.plotting.plot_xy(df, interactive=False)


def test_polar_bar():
    traja.plotting.polar_bar(df, interactive=False)


def test_find_runs():
    actual = traja.find_runs(df.x)
    expected = (
        np.array(
            [
                0.0,
                1.323_370_69,
                2.275_837_54,
                2.274_285_61,
                0.336_029_88,
                -1.455_690_92,
                -3.544_442_11,
                -5.386_597_93,
                -7.508_544_4,
                -9.353_255_17,
            ]
        ),
        np.array([0, 1, 2, 3, 4, 5, 6, 7, 8, 9]),
        np.array([1, 1, 1, 1, 1, 1, 1, 1, 1, 1]),
    )
    for i in range(len(actual)):
        npt.assert_allclose(actual[i], expected[i])


def test_plot_clustermap():
    trjs = [traja.generate(seed=i) for i in range(20)]

    # Calculate displacement
    displacements = [trj.traja.calc_displacement() for trj in trjs]

    traja.plot_clustermap(displacements)


def test_plot():
    ax = traja.plotting.plot(df, interactive=False)
    path = ax.get_paths()[0]
    npt.assert_allclose(
        path._vertices[:5],
        np.array(
            [
                [0.0, -0.5],
                [0.132_601_55, -0.5],
                [0.259_789_94, -0.447_316_85],
                [0.353_553_39, -0.353_553_39],
                [0.447_316_85, -0.259_789_94],
            ]
        ),
    )



# class Seq2Seq(nn.Module):
#
#     def __init__(self, input_size=2, hidden_size=100, output_size=2):
#         super(Seq2Seq, self).__init__()
#         self.encoder_rnn = EncoderRNN(input_size, hidden_size)
#         self.decoder_rnn = DecoderRNN(hidden_size, output_size)
#
#     def forward(self, input, hidden):
#         bottleneck = self.encoder_rnn(input, hidden)
#         pred = self.decoder_rnn(input, hidden)
#
#         return pred
#
#
# def train_seq2seq(input_tensor, target_tensor, encoder, decoder, encoder_optimizer, decoder_optimizer, criterion,
#           max_length=nb_steps):
#     encoder_hidden = encoder.initHidden()
#
#     encoder_optimizer.zero_grad()
#     decoder_optimizer.zero_grad()
#
#     input_length = input_tensor.size(0)
#     target_length = target_tensor.size(0)
#
#     encoder_outputs = torch.zeros(max_length, encoder.hidden_size, device=device)
#
#     loss = 0
#
#     for ei in range(input_length):
#         encoder_output, encoder_hidden = encoder(
#             input_tensor[ei], encoder_hidden)
#         encoder_outputs[ei] = encoder_output[0, 0]
#
#     decoder_input = torch.tensor([[SOS_token]], device=device)
#
#     decoder_hidden = encoder_hidden
#
#
#     # Without teacher forcing: use its own predictions as the next input
#     for di in range(target_length):
#         decoder_output, decoder_hidden, decoder_attention = decoder(
#             decoder_input, decoder_hidden, encoder_outputs)
#         topv, topi = decoder_output.topk(1)
#         decoder_input = topi.squeeze().detach()  # detach from history as input
#
#         loss += criterion(decoder_output, target_tensor[di])
#
#     loss.backward()
#
#     encoder_optimizer.step()
#     decoder_optimizer.step()
#
#     return loss.item() / target_length
#
# import time
# import math
#
#
# def asMinutes(s):
#     m = math.floor(s / 60)
#     s -= m * 60
#     return '%dm %ds' % (m, s)
#
#
# def timeSince(since, percent):
#     now = time.time()
#     s = now - since
#     es = s / (percent)
#     rs = es - s
#     return '%s (- %s)' % (asMinutes(s), asMinutes(rs))
#
#
# def trainIters_seq2sec(encoder, decoder, n_iters, print_every=1000, plot_every=100, learning_rate=0.01):
#     import time
#     start = time.time()
#     plot_losses = []
#     print_loss_total = 0  # Reset every print_every
#     plot_loss_total = 0  # Reset every plot_every
#
#     encoder_optimizer = optim.SGD(encoder.parameters(), lr=learning_rate)
#     decoder_optimizer = optim.SGD(decoder.parameters(), lr=learning_rate)
#     training_pairs = [tensorsFromPair(random.choice(pairs))
#                       for i in range(n_iters)]
#     criterion = nn.NLLLoss()
#
#     for iter in range(1, n_iters + 1):
#         training_pair = training_pairs[iter - 1]
#         input_tensor = training_pair[0]
#         target_tensor = training_pair[1]
#
#         loss = train(input_tensor, target_tensor, encoder,
#                      decoder, encoder_optimizer, decoder_optimizer, criterion)
#         print_loss_total += loss
#         plot_loss_total += loss
#
#         if iter % print_every == 0:
#             print_loss_avg = print_loss_total / print_every
#             print_loss_total = 0
#             print('%s (%d %d%%) %.4f' % (timeSince(start, iter / n_iters),
#                                          iter, iter / n_iters * 100, print_loss_avg))
#
#         if iter % plot_every == 0:
#             plot_loss_avg = plot_loss_total / plot_every
#             plot_losses.append(plot_loss_avg)
#             plot_loss_total = 0
#
#     # showPlot(plot_losses)


class Seq2seq(nn.Module):
    def __init__(
        self,
        input_size=2,
        hidden_size=100,
        output_size=2,
        decode_function=F.log_softmax,
    ):
        super(Seq2seq, self).__init__()
        self.encoder = encoder
        self.decoder = decoder
        self.decode_function = decode_function

    def forward(self, input, hidden):
        bottleneck = self.encoder_rnn(input, hidden)
        pred = self.decoder_rnn(input, hidden)

        return pred

    def flatten_parameters(self):
        self.encoder.rnn.flatten_parameters()
        self.decoder.rnn.flatten_parameters()

    def forward(
        self,
        input_variable,
        input_lengths=None,
        target_variable=None,
        teacher_forcing_ratio=0,
    ):
        encoder_outputs, encoder_hidden = self.encoder(input_variable, input_lengths)
        result = self.decoder(
            input=target_variable,
            encoder_hidden=encoder_hidden,
            encoder_outputs=encoder_outputs,
            function=self.decode_function,
            teacher_forcing_ratio=teacher_forcing_ratio,
        )
        return result


class EncoderRNN(nn.Module):
    def __init__(self, input_size, hidden_size):
        super(EncoderRNN, self).__init__()
        self.hidden_size = hidden_size

        self.embedding = nn.Embedding(input_size, hidden_size)
        self.gru = nn.GRU(hidden_size, hidden_size)

    def forward(self, input, hidden):
        embedded = self.embedding(input).view(1, 1, -1)
        output = embedded
        output, hidden = self.gru(output, hidden)
        return output, hidden

    def initHidden(self):
        return torch.zeros(1, 1, self.hidden_size, device=device)


class DecoderRNN(nn.Module):
    def __init__(self, hidden_size, output_size):
        super(DecoderRNN, self).__init__()
        self.hidden_size = hidden_size

        self.embedding = nn.Embedding(output_size, hidden_size)
        self.gru = nn.GRU(hidden_size, hidden_size)
        self.out = nn.Linear(hidden_size, output_size)
        self.softmax = nn.LogSoftmax(dim=1)

    def forward(self, input, hidden):
        output = self.embedding(input).view(1, 1, -1)
        output = F.relu(output)
        output, hidden = self.gru(output, hidden)
        output = self.softmax(self.out(output[0]))
        return output, hidden

    def initHidden(self):
        return torch.zeros(1, 1, self.hidden_size, device=device)
